"""
Utilities for fitting stacked (drizzled) spectra
"""
from collections import OrderedDict

import astropy.io.fits as pyfits
import astropy.units as u

import numpy as np

from . import utils
from .utils import GRISM_COLORS, GRISM_MAJOR, GRISM_LIMITS, DEFAULT_LINE_LIST

from .fitting import GroupFitter


def make_templates(
    grism="G141", return_lists=False, fsps_templates=False, line_list=DEFAULT_LINE_LIST
):
    """
    Generate template savefile

    This script generates the template sets with the emission line
    complexes and with individual lines.

    Parameters
    ----------
    grism : str
        Grism of interest, which defines what FWHM to use for the line
        templates.

    return_lists : bool
        Return the templates rather than saving them to a file

    fsps_templates : bool
        Use FSPS templates instead of line complex templates

    line_list : list
        List of individual lines to include in the templates

    Returns
    -------
    t_complexes, t_lines : list
        If `return` then return two lists of templates.  Otherwise,
        store them to a `~numpy` save file "templates_{fwhm}.npy".

    """

    from .multifit import MultiBeam

    if grism == "G141":  # WFC3/IR
        fwhm = 1100
    elif grism == "G800L":  # ACS/UVIS
        fwhm = 1400
    elif grism == "G280":  # WFC3/UVIS
        fwhm = 1500
    elif grism == "GRISM":  # WFIRST
        fwhm = 350
    else:
        fwhm = 700  # G102

    # Line complex templates
    t_complexes = utils.load_templates(
        fwhm=fwhm, line_complexes=True, fsps_templates=fsps_templates
    )

    # Individual lines
    # line_list = ['SIII', 'SII', 'Ha', 'OI-6302', 'OIII', 'Hb',
    #              'OIII-4363', 'Hg', 'Hd', 'NeIII', 'OII', 'MgII']

    t_lines = utils.load_templates(
        fwhm=fwhm,
        line_complexes=False,
        full_line_list=line_list,
        fsps_templates=fsps_templates,
    )

    if return_lists:
        return t_complexes, t_lines
    else:
        # Save them to a file
        np.save("templates_{0}.npy".format(fwhm), [t_complexes, t_lines])
        print("Wrote `templates_{0}.npy`".format(fwhm))


class StackFitter(GroupFitter):
    def __init__(
        self,
        files="gnt_18197.stack.fits",
        group_name=None,
        sys_err=0.02,
        mask_min=0.1,
        fit_stacks=True,
        fcontam=1,
        PAs=None,
        extensions=None,
        min_ivar=0.01,
        overlap_threshold=3,
        verbose=True,
        eazyp=None,
        eazy_ix=0,
        MW_EBV=0.0,
        chi2_threshold=1.5,
        min_DoF=200,
    ):
        """
        Object for fitting stacked spectra.

        Parameters
        ----------
        files : str or list of str
            Stack FITS filename.  If a list is supplied, e.g., the product
            of a `~glob` command, then append all specified files.

        group_name : str
            Rootname to associate with the object.  If none, then default to
            `files`.

        sys_err : float
            Minimum systematic error, interpreted as a fractional error.
            The adjusted variance is taken to be

                >>> var = var0 + (sys_err*flux)**2

        mask_min : float
            Only fit 2D pixels where the flat-flambda model has pixel values
            greater than `mask_min` times the maximum of the model.

        fit_stacks : bool
            Fit the stacks of each grism combined from all available PAs.  If
            False, then fit the PAs individually.

        fcontam : float
            Parameter to control weighting of contaminated pixels for
            `fit_stacks=False`.

        PAs : list of int
            List of PAs to fit. If None, fit all PAs.

        extensions : list of str
            List of extensions to fit. If None, fit all extensions.

        min_ivar : float
            Minimum inverse variance threshold for fitting.
            Default = 0.01.

        overlap_threshold : float
            Threshold for masking overlapping regions.
            Default = 3.

        verbose : bool
            If True, print verbose output.

        eazyp : `~eazy.photoz.EazyPhot`
            EazyPhot object for photometric redshift fitting.

        eazy_ix : int
            Index of the EazyPhot object to use for photometric redshift fitting.
            Default = 0.

        MW_EBV : float
            Milky Way E(B-V) reddening value.
            Default = 0.

        chi2_threshold : float
            Threshold for identifying bad PAs based on chi-squared values.
            Default = 1.5 . 

        min_DoF : int
            Minimum degrees of freedom threshold for identifying bad PAs.
            Default = 200.

        """
        if isinstance(files, list):
            file = files[0]
        else:
            file = files

        self.files = [file]
        if group_name is not None:
            self.group_name = group_name
        else:
            self.group_name = file

        if verbose:
            print("Load file {0}".format(file))

        self.file = file
        self.hdulist = pyfits.open(file)
        self.min_ivar = min_ivar
        self.sys_err = sys_err
        self.fcontam = fcontam

        self.MW_EBV = MW_EBV

        self.h0 = self.hdulist[0].header.copy()
        # self.Ngrism = self.h0['NGRISM']
        self.grisms = []
        self.ext = []
        for i in range(self.h0["NGRISM"]):
            g = self.h0["GRISM{0:03d}".format(i + 1)]
            self.grisms.append(g)
            if fit_stacks:
                if extensions is not None:
                    if g not in extensions:
                        continue

                self.ext.append(g)
            else:
                ng = self.h0["N{0}".format(g)]
                for j in range(ng):
                    pa = self.h0["{0}{1:02d}".format(g, j + 1)]

                    if PAs is not None:
                        if pa not in PAs:
                            continue

                    ext = "{0},{1}".format(g, pa)
                    if extensions is not None:
                        if ext not in extensions:
                            continue

                    self.ext.append(ext)

        self.N = len(self.ext)
        self.beams = []
        pop = []
        for i in range(self.N):
            E_i = StackedSpectrum(
                file=self.file,
                sys_err=sys_err,
                mask_min=mask_min,
                extver=self.ext[i],
                mask_threshold=-1,
                fcontam=fcontam,
                min_ivar=min_ivar,
                MW_EBV=MW_EBV,
            )
            E_i.compute_model()

            if np.isfinite(E_i.kernel.sum()) & (E_i.DoF >= min_DoF):
                self.beams.append(E_i)
            else:
                pop.append(i)

        for i in pop[::-1]:
            self.N -= 1
            p = self.ext.pop(i)

        # Get some parameters from the beams
        self.id = self.h0["ID"]
        self.ra = self.h0["RA"]
        self.dec = self.h0["DEC"]

        self.Asave = {}

        # Photometry
        self.is_spec = 1
        self.Nphot = 0

        # Parse the beam data
        self._parse_beams_list()

        if not fit_stacks:
            # self.mask_drizzle_overlaps(threshold=overlap_threshold,
            #                           verbose=verbose)
            if chi2_threshold > 0:
                orig_ext = [e for e in self.ext]
                fit_log, keep_dict, has_bad = self.check_for_bad_PAs(
                    poly_order=3,
                    chi2_threshold=chi2_threshold,
                    fit_background=True,
                    reinit=True,
                    verbose=False,
                )
                if has_bad & verbose:
                    print("Found bad PA.  New list: {0}".format(keep_dict))

        if verbose:
            print("  {0}".format(" ".join(self.ext)))

        # Read multiple
        if isinstance(files, list):
            if len(files) > 1:
                for file in files[1:]:
                    extra = StackFitter(
                        files=file,
                        sys_err=sys_err,
                        mask_min=mask_min,
                        fit_stacks=fit_stacks,
                        fcontam=fcontam,
                        pas=PAs,
                        extensions=extensions,
                        min_ivar=min_ivar,
                        overlap_threshold=overlap_threshold,
                        eazyp=eazyp,
                        eazy_ix=eazy_ix,
                        chi2_threshold=chi2_threshold,
                        verbose=verbose,
                    )
                    self.extend(extra)

        self.idf = np.hstack([b.scif * 0 + ib for ib, b in enumerate(self.beams)])
        self.idf = np.asarray(self.idf, dtype=int)

        # if eazyp is not None:
        #     self.eazyp = eazyp
        #
        #     # TBD: do matching to eazyp.cat directly?
        #     self.eazy_ix = eazy_ix
        #
        #     ok_phot = (eazyp.efnu[eazy_ix,:] > 0) & (eazyp.fnu[eazy_ix,:] > eazyp.param['NOT_OBS_THRESHOLD']) & np.isfinite(eazyp.fnu[eazy_ix,:]) & np.isfinite(eazyp.efnu[eazy_ix,:])
        #     ok_phot = np.squeeze(ok_phot)
        #     self.ok_phot = ok_phot
        #
        #     self.Nphot = ok_phot.sum()
        #     if self.Nphot > 0:
        #
        #         # F-lambda photometry, 1e-19 erg/s/cm2/A
        #         self.photom_eflam = (eazyp.efnu[eazy_ix,:]*eazyp.to_flam*eazyp.zp*eazyp.ext_corr/100.)[ok_phot]
        #         self.photom_flam = (eazyp.fnu[eazy_ix,:]*eazyp.to_flam*eazyp.zp*eazyp.ext_corr/100.)[ok_phot]
        #         self.photom_lc = eazyp.lc[ok_phot]
        #
        #         self.scif = np.hstack((self.scif, self.photom_flam))
        #         self.ivarf = np.hstack((self.ivarf, 1/self.photom_eflam**2))
        #         self.sivarf = np.hstack((self.sivarf, 1/self.photom_eflam))
        #         self.wavef = np.hstack((self.wavef, self.photom_lc))
        #
        #         self.weightf = np.hstack((self.weightf, np.ones(self.Nphot)))
        #         self.fit_mask = np.hstack((self.fit_mask, np.ones(self.Nphot, dtype=bool)))
        #         self.DoF += self.Nphot
        #         self.phot_scale = np.array([10.])

    def _parse_beams_list(self):
        """
        Parses the `self.beams` list and extracts various information such 
        as the number of beams, the extension versions and other attributes.
        
        """

        self.N = len(self.beams)
        self.ext = [E.extver for E in self.beams]

        self.Ngrism = OrderedDict()
        for beam in self.beams:
            if beam.grism in self.Ngrism:
                self.Ngrism[beam.grism] += 1
            else:
                self.Ngrism[beam.grism] = 1

        # Make "PA" attribute
        self.PA = OrderedDict()
        for g in self.Ngrism:
            self.PA[g] = OrderedDict()

        for i in range(self.N):
            grism = self.ext[i].split(",")[0]
            if "," in self.ext[i]:
                PA = float(self.ext[i].split(",")[1])
            else:
                PA = 0

            if PA in self.PA[grism]:
                self.PA[grism][PA].append(i)
            else:
                self.PA[grism][PA] = [i]

        self.grisms = list(self.PA.keys())

        self.Ntot = np.sum([E.size for E in self.beams])
        self.scif = np.hstack([E.scif for E in self.beams])
        self.ivarf = np.hstack([E.ivarf for E in self.beams])
        self.wavef = np.hstack([E.wavef for E in self.beams])

        self.weightf = np.hstack([E.weightf for E in self.beams])
        # self.ivarf *= self.weightf

        self.sivarf = np.sqrt(self.ivarf)

        self.fit_mask = np.hstack([E.fit_mask for E in self.beams])
        self.fit_mask &= self.ivarf > self.min_ivar * self.ivarf.max()

        # Dummy parameter.  Implemented for MultiBeam
        self.sens_mask = 1.0

        self.DoF = int((self.fit_mask * self.weightf).sum())
        # self.Nmask = self.fit_mask.sum()
        self.Nmask = np.sum([E.fit_mask.sum() for E in self.beams])

        self.slices = self._get_slices(masked=False)
        self.A_bg = self._init_background(masked=False)
        self.Asave = {}

        self._update_beam_mask()
        self.A_bgm = self._init_background(masked=True)

        self.flat_flam = np.hstack([E.flat_flam for E in self.beams])

        self.initialize_masked_arrays()

    def extend(self, st):
        """
        Append second StackFitter objects to `self`.

        Parameters
        ----------
        st : StackFitter
            The StackFitter object to be appended.

        """

        self.beams.extend(st.beams)
        self._parse_beams_list()

        # self.grisms.extend(st.grisms)
        # self.grisms = list(np.unique(self.grisms))
        #
        # self.Ngrism = {}
        # for grism in self.grisms:
        #     self.Ngrism[grism] = 0
        #
        # for beam in self.beams:
        #     self.Ngrism[beam.grism] += 1
        #
        # #self.Ngrism = len(self.grisms)
        #
        # self.N += st.N
        # self.ext.extend(st.ext)
        # self.files.extend(st.files)
        #
        # # Re-init
        # self.Ntot = np.sum([E.size for E in self.beams])
        # self.scif = np.hstack([E.scif for E in self.beams])
        # self.ivarf = np.hstack([E.ivarf for E in self.beams])
        # self.wavef = np.hstack([E.wavef for E in self.beams])
        #
        # self.weightf = np.hstack([E.weightf for E in self.beams])
        # #self.ivarf *= self.weightf
        #
        # self.sivarf = np.sqrt(self.ivarf)
        #
        # self.fit_mask = np.hstack([E.fit_mask for E in self.beams])
        # self.fit_mask &= self.ivarf > self.min_ivar*self.ivarf.max()
        #
        # self.slices = self._get_slices(masked=False)
        # self.A_bg = self._init_background(masked=False)
        #
        # self._update_beam_mask()
        # self.A_bgm = self._init_background(masked=True)
        #
        # self.Nmask = self.fit_mask.sum()
        # self.DoF = int((self.fit_mask*self.weightf).sum())
        #
        # self.flat_flam = np.hstack([E.flat_flam for E in self.beams])


    def check_for_bad_PAs(
        self,
        poly_order=1,
        chi2_threshold=1.5,
        fit_background=True,
        reinit=True,
        verbose=False,
    ):
        """
        Check for bad PAs and return fit log, keep dictionary, and 
        a flag indicating if there are bad PAs.

        Parameters
        ----------
        poly_order : int
            The order of the polynomial templates used for fitting. 
            Default is 1.

        chi2_threshold : float
            The threshold value for the chi-square ratio below which a PA 
            is considered good. Default is 1.5.

        fit_background : bool
            Whether to fit the background when calculating the chi-square ratio. 
            Default is True.

        reinit : bool
            Whether to reinitialize the `self` object with the beams 
            corresponding to the good PAs. Default is True.

        verbose : bool
            Whether to print verbose output. Default is False.

        Returns
        -------
        fit_log : OrderedDict
            A dictionary containing the chi-square values, degrees of freedom, 
            and chi-square ratios for each PA and group.

        keep_dict : OrderedDict
            A dictionary containing the list of PAs to keep for each group.
            
        has_bad : bool
            A flag indicating if there are bad PAs.

        """

        wave = np.linspace(2000, 2.5e4, 100)
        poly_templates = utils.polynomial_templates(wave, order=poly_order)

        fit_log = OrderedDict()
        keep_dict = OrderedDict()
        has_bad = False

        keep_beams = []

        for g in self.PA:
            fit_log[g] = OrderedDict()
            keep_dict[g] = []

            for pa in self.PA[g]:
                extensions = [self.ext[i] for i in self.PA[g][pa]]
                mb_i = StackFitter(
                    self.file,
                    fcontam=self.fcontam,
                    sys_err=self.sys_err,
                    extensions=extensions,
                    fit_stacks=False,
                    verbose=verbose,
                    chi2_threshold=-1,
                )

                try:
                    chi2, _, _, _ = mb_i.xfit_at_z(
                        z=0, templates=poly_templates, fit_background=fit_background
                    )
                except:
                    chi2 = 1e30

                if False:
                    p_i = mb_i.template_at_z(
                        z=0,
                        templates=poly_templates,
                        fit_background=fit_background,
                        fitter="lstsq",
                        fwhm=1400,
                        get_uncertainties=2,
                    )

                fit_log[g][pa] = {
                    "chi2": chi2,
                    "DoF": mb_i.DoF,
                    "chi_nu": chi2 / np.maximum(mb_i.DoF, 1),
                }

            min_chinu = 1e30
            for pa in self.PA[g]:
                min_chinu = np.minimum(min_chinu, fit_log[g][pa]["chi_nu"])

            fit_log[g]["min_chinu"] = min_chinu

            for pa in self.PA[g]:
                fit_log[g][pa]["chinu_ratio"] = fit_log[g][pa]["chi_nu"] / min_chinu

                if fit_log[g][pa]["chinu_ratio"] < chi2_threshold:
                    keep_dict[g].append(pa)
                    keep_beams.extend([self.beams[i] for i in self.PA[g][pa]])
                else:
                    has_bad = True

        if reinit:
            self.beams = keep_beams
            self._parse_beams_list()
            # self._parse_beams(psf=self.psf_param_dict is not None)

        return fit_log, keep_dict, has_bad

    def compute_model(self, spectrum_1d=None):
        """
        TBD
        """
        return False

    def get_sky_coords(self):
        """
        Get WCS coordinates of the center of the direct image

        Returns
        -------
        ra, dec : float
            Center coordinates of the beam thumbnail in decimal degrees
        """
        ra = self.beams[0].h0["RA"]
        dec = self.beams[0].h0["DEC"]
        return ra, dec

    def fit_combined_at_z(
        self,
        z=0,
        fitter="nnls",
        get_uncertainties=False,
        eazyp=None,
        ix=0,
        order=1,
        scale_fit=None,
        method="BFGS",
    ):
        """
        Fit the 2D spectra with a set of templates at a specified redshift.

        Parameters
        ----------
        z : float
            Redshift. Default is 0.

        fitter : str
            Minimization algorithm to compute template coefficients.
            The default 'nnls' uses non-negative least squares.
            The other option is standard 'leastsq'.

        get_uncertainties : bool
            Compute coefficient uncertainties from the covariance matrix.
            Default is False.

        eazyp : `~eazy.photoz.EazyPhot`
            EazyPhot object containing the EAZY templates and filter information.
            Default in None.

        ix : int
            Index of the object in the EazyPhot object.
            Default is 0.

        order : int
            The order of the polynomial templates used for fitting. 
            Default is 1.

        scale_fit : `~scipy.optimize.OptimizeResult`
            Object containing the result of the optimization process.
            Default is None.

        method : str
            The optimization method to use. Default is 'BFGS'.

        Returns
        -------
        chi2 : float
            Chi-squared of the fit

        background : `~np.ndarray`
            Background model

        full : `~np.ndarray`
            Full model

        full_coeffs : `~np.ndarray`
            Coefficients of the templates

        full_coeffs_err : `~np.ndarray`
            Uncertainties of the template coefficients

        scale_fit : `~scipy.optimize.OptimizeResult`
            Object containing the result of the optimization process.

        """
        import scipy.optimize

        # Photometry
        ok_phot = (
            (eazyp.efnu[ix, :] > 0)
            & (eazyp.fnu[ix, :] > eazyp.param["NOT_OBS_THRESHOLD"])
            & np.isfinite(eazyp.fnu[ix, :])
            & np.isfinite(eazyp.efnu[ix, :])
        )
        ok_phot = np.squeeze(ok_phot)
        self.ok_phot = ok_phot

        Nphot = ok_phot.sum()
        photom_eflam = (eazyp.efnu[ix, :] * eazyp.to_flam * eazyp.ext_corr / 100.0)[
            ok_phot
        ]
        photom_flam = (eazyp.fnu[ix, :] * eazyp.to_flam * eazyp.ext_corr / 100.0)[
            ok_phot
        ]

        templates = eazyp.templates

        NTEMP = len(templates)
        A = np.zeros((self.N + NTEMP, self.Ntot + Nphot))
        A[: self.N, :-Nphot] += self.A_bg

        pedestal = 0.04

        sivarf = np.hstack((self.sivarf, 1 / photom_eflam))
        dataf = np.hstack((self.scif + pedestal, photom_flam))
        fit_mask = np.hstack((self.fit_mask, np.ones(Nphot, dtype=bool)))
        # sivarf *= fit_mask

        # Photometry
        Aphot = (eazyp.tempfilt(z) * 3.0e18 / eazyp.lc ** 2 * (1 + z))[:, ok_phot]
        A[self.N :, -Nphot:] += Aphot

        for i, ti in enumerate(templates):
            # ti = templates[t]
            s = [ti.wave * (1 + z), ti.flux / (1 + z)]

            for j, E in enumerate(self.beams):
                clip = E.ivar.sum(axis=0) > 0
                if (s[0][0] > E.wave[clip].max()) | (s[0][-1] < E.wave[clip].min()):
                    continue

                sl = self.slices[j]
                A[self.N + i, sl] = E.compute_model(spectrum_1d=s)

        oktemp = (A * fit_mask).sum(axis=1) != 0

        # ATA = np.dot(A[oktemp,:]*(sivarf*fit_mask)**2, A[oktemp,:].T)
        # ATy = np.dot(A[oktemp,:]*(sivarf*fit_mask)**2, dataf)

        Ax = A[oktemp, :] * sivarf
        # AxT = Ax[:,fit_mask].T
        # data = (dataf*sivarf)[fit_mask]

        # Run the optimizer
        # method = 'Powell'
        # method = 'BFGS'
        tol = 1.0e-4
        init = np.zeros(order + 1)
        init[0] = 10.0

        if scale_fit is None:
            scale_fit = scipy.optimize.minimize(
                self.objective_scale,
                init,
                args=(
                    Ax,
                    dataf * sivarf,
                    self.wavef,
                    fit_mask,
                    sivarf,
                    Nphot,
                    self.N,
                    0,
                ),
                method=method,
                jac=None,
                hess=None,
                hessp=None,
                bounds=(),
                constraints=(),
                tol=tol,
                callback=None,
                options=None,
            )

            if order == 0:
                scale_fit.x = np.array([float(scale_fit.x)])

        coeffs, full, resid, chi2, AxT = self.objective_scale(
            scale_fit.x,
            Ax,
            dataf * sivarf,
            self.wavef,
            fit_mask,
            sivarf,
            Nphot,
            self.N,
            True,
        )

        background = np.dot(coeffs[: self.N], A[: self.N, :])
        full -= background

        background -= pedestal
        background[-Nphot:] = 0

        # Uncertainties from covariance matrix
        if get_uncertainties:
            try:
                # covar = np.matrix(np.dot(AxT.T, AxT)).I
                covar = utils.safe_invert(np.dot(AxT.T, AxT))
                covard = np.sqrt(covar.diagonal()).A.flatten()

                # covarf = np.matrix(np.dot(ATA.T, ATA)).I
                # covardf = np.sqrt(covarf.diagonal()).A.flatten()

            except:
                print("Except!")
                covard = np.zeros(oktemp.sum())  # -1.
        else:
            covard = np.zeros(oktemp.sum())  # -1.

        full_coeffs = np.zeros(NTEMP)
        full_coeffs[oktemp[self.N :]] = coeffs[self.N :]

        full_coeffs_err = np.zeros(NTEMP)
        full_coeffs_err[oktemp[self.N :]] = covard[self.N :]

        return chi2, background, full, full_coeffs, full_coeffs_err, scale_fit

    @staticmethod
    def scale_AxT(p, Ax, spec_wave, Nphot, Next):
        """
        Scale spectrum templates by polynomial function.

        Parameters
        ----------
        p : array_like
            The coefficients of the polynomial function.

        Ax : array_like
            The scaled template matrix.

        spec_wave : array_like
            The wavelength array.

        Nphot : int
            The number of photometric points.

        Next : int
            The number of template components.

        Returns
        -------
        AxT : array_like
            The scaled template matrix.

        """
        from np.polynomial import Polynomial

        scale = np.ones(Ax.shape[1])
        scale[:-Nphot] = Polynomial(p / 10.0)((spec_wave - 1.0e4) / 1000.0)
        AxT = Ax * scale
        for i in range(Next):
            AxT[i, :] /= scale

        return AxT

    @staticmethod
    def objective_scale(
            p, Ax, data, spec_wave, fit_mask, sivarf, Nphot, Next, return_coeffs
        ):
        """
        Objective function for fitting for a scale term between photometry and
        spectra

        Parameters
        ----------
        p : float
            Scale term between photometry and spectra.

        Ax : `~np.ndarray`
            Array of shape (M, N) representing the design matrix.

        data : `~np.ndarray`
            Array of shape (N,) representing the observed data.

        spec_wave : `~np.ndarray`
            Array of shape (N,) representing the wavelength of the spectra.

        fit_mask : `~np.ndarray`
            Boolean array of shape (N,) representing the mask for fitting.

        sivarf : `~np.ndarray`
            Array of shape (N,) representing the inverse variance of the spectra.

        Nphot : int
            Number of photometric data points.

        Next : int
            Number of background components.
            
        return_coeffs : bool
            Flag indicating whether to return the coefficients.

        Returns
        -------
        if return_coeffs:
        
            coeffs : `~np.ndarray`
                Array of shape (M,) representing the coefficients of the fit.

            full : `~np.ndarray`
                Array of shape (N,) representing the full model.
            
            resid : `~np.ndarray`
                Array of shape (N,) representing the residuals.

            chi2 : float
                The chi-squared value of the fit.
            
            AxT : `~np.ndarray`
                The scaled template matrix.
        
        else:
        
            chi2 : float
                The chi-squared value of the fit.

        """
        import scipy.optimize
        from np.polynomial import Polynomial

        scale = np.ones(Ax.shape[1])
        scale[:-Nphot] = Polynomial(p / 10.0)((spec_wave - 1.0e4) / 1000.0)
        AxT = Ax * scale
        # Remove scaling from background component
        for i in range(Next):
            AxT[i, :] /= scale

        #AxT = AxT[:,fit_mask].T
        # (Ax*scale)[:,fit_mask].T
        #AxT[:,:Next] = 1.

        coeffs, rnorm = scipy.optimize.nnls(AxT[:, fit_mask].T, data[fit_mask])
        full = np.dot(coeffs, AxT / sivarf)
        resid = data / sivarf - full  # - background
        chi2 = np.sum(resid[fit_mask] ** 2 * sivarf[fit_mask] ** 2)
        if return_coeffs:
            return coeffs, full, resid, chi2, AxT
        else:
            return chi2
        
    def fit_at_z(self, z=0, templates=[], fitter="nnls", get_uncertainties=False):
        """
        Fit the 2D spectra with a set of templates at a specified redshift.

        Parameters
        ----------
        z : float
            Redshift.

        templates : list
            List of templates to fit.

        fitter : str
            Minimization algorithm to compute template coefficients.
            The default 'nnls' uses non-negative least squares.
            The other option is standard 'leastsq'.

        get_uncertainties : bool
            Compute coefficient uncertainties from the covariance matrix


        Returns
        -------
        chi2 : float
            Chi-squared of the fit

        background : `~np.ndarray`
            Background model

        full_model : `~np.ndarray`
            Best-fit 2D model.

        coeffs, err : `~np.ndarray`
            Template coefficients and uncertainties.

        """
        import scipy.optimize

        NTEMP = len(templates)
        A = np.zeros((self.N + NTEMP, self.Ntot))
        A[: self.N, :] += self.A_bg

        for i, t in enumerate(templates):
            ti = templates[t]
            try:
                import eazy.igm

                if z > 7:
                    igm = eazy.igm.Inoue14()
                    igmz = igm.full_IGM(z, ti.wave * (1 + z))
                else:
                    igmz = 1.0

            except:
                igmz = 1.0

            s = [ti.wave * (1 + z), ti.flux / (1 + z) * igmz]

            for j, E in enumerate(self.beams):
                clip = E.ivar.sum(axis=0) > 0
                if (s[0][0] > E.wave[clip].max()) | (s[0][-1] < E.wave[clip].min()):
                    continue

                sl = self.slices[j]
                A[self.N + i, sl] = E.compute_model(spectrum_1d=s)

        oktemp = (A * self.fit_mask).sum(axis=1) != 0

        Ax = A[oktemp, :] * self.sivarf

        pedestal = 0.04

        AxT = Ax[:, self.fit_mask].T
        data = ((self.scif + pedestal) * self.sivarf)[self.fit_mask]

        if fitter == "nnls":
            coeffs, rnorm = scipy.optimize.nnls(AxT, data)
        else:
            coeffs, residuals, rank, s = np.linalg.lstsq(AxT, data)

        background = np.dot(coeffs[: self.N], A[: self.N, :]) - pedestal
        full = np.dot(coeffs[self.N :], Ax[self.N :, :] / self.sivarf)

        resid = self.scif - full - background
        chi2 = np.sum(resid[self.fit_mask] ** 2 * self.sivarf[self.fit_mask] ** 2)

        # Uncertainties from covariance matrix
        if get_uncertainties:
            try:
                # covar = np.matrix(np.dot(AxT.T, AxT)).I
                covar = utils.safe_invert(np.dot(AxT.T, AxT))
                covard = np.sqrt(covar.diagonal()).A.flatten()
            except:
                print("Except: covar!")
                covard = np.zeros(oktemp.sum())  # -1.
        else:
            covard = np.zeros(oktemp.sum())  # -1.

        full_coeffs = np.zeros(NTEMP)
        full_coeffs[oktemp[self.N :]] = coeffs[self.N :]

        full_coeffs_err = np.zeros(NTEMP)
        full_coeffs_err[oktemp[self.N :]] = covard[self.N :]

        return chi2, background, full, full_coeffs, full_coeffs_err

    def fit_zgrid(
        self,
        dz0=0.005,
        zr=[0.4, 3.4],
        fitter="nnls",
        make_plot=True,
        save_data=True,
        prior=None,
        templates_file="templates.npy",
        verbose=True,
        outlier_threshold=1e30,
        eazyp=None,
        ix=0,
        order=0,
        scale_fit=None,
    ):
        """
        Fit templates on a redshift grid.

        Parameters
        ----------
        dz0 : float
            Initial step size of the redshift grid (dz/1+z).
            Default is 0.005.

        zr : list
            Redshift range to consider.
            Default is [0.4, 3.4].

        fitter : str
            Minimization algorithm. Default is non-negative least-squares.
            Default is 'nnls'.

        make_plot : bool
            Make the diagnostic plot. Default is True.

        prior : list
            Naive prior to add to the nominal chi-squared(z) of the template
            fits. The example below is a simple Gaussian prior centered
            at z=1.5.

            >>> z_prior = np.arange(0,3,0.001)
            >>> chi_prior = (z_prior-1.5)**2/2/0.1**2
            >>> prior = [z_prior, chi_prior]

        templates_file : str
            Filename of the `~numpy` save file containing the templates. Use
            the `make_templates` script to generate this. 
            Default is 'templates.npy'.

        verbose : bool
            Print the redshift grid steps. Default is True.

        outlier_threshold : float
            Threshold for masking outliers in the fit. Default is 1e30.

        eazyp : `~eazy.photoz.EazyTemplate`
            EazyTemplate object for combined fits. Default is None.

        ix : int
            Index of the object in the eazyp object. Default is 0.

        order : int
            Order of the polynomial fit to the combined spectrum. Default is 0.

        scale_fit : float
            Scale factor for the combined spectrum. Default is None.

        Returns
        -------
        hdu : `~astropy.io.fits.HDUList`
            Multi-extension FITS file with the result of the redshift fits.

        """
        import os

        # import grizli

        import matplotlib.gridspec
        import matplotlib.pyplot as plt
        import numpy as np

        from . import utils

        t_complex, t_i = np.load(templates_file, allow_pickle=True)

        z = utils.log_zgrid(zr=zr, dz=dz0)
        chi2 = z * 0.0
        for i in range(len(z)):
            if eazyp:
                out = self.fit_combined_at_z(
                    z=z[i], eazyp=eazyp, ix=ix, order=order, scale_fit=scale_fit
                )
                chi2[i], bg, full, coeffs, err, scale_fit = out
            else:
                out = self.fit_at_z(z=z[i], templates=t_complex)
                chi2[i], bg, full, coeffs, err = out

            if verbose:
                print("{0:.4f} - {1:10.1f}".format(z[i], chi2[i]))

        # Zoom in on the chi-sq minimum.
        ci = chi2
        zi = z
        for iter in range(1, 7):
            if prior is not None:
                pz = np.interp(zi, prior[0], prior[1])
                cp = ci + pz
            else:
                cp = ci

            iz = np.argmin(cp)
            z0 = zi[iz]
            dz = dz0 / 2.02 ** iter
            zi = utils.log_zgrid(zr=[z0 - dz * 4, z0 + dz * 4], dz=dz)
            ci = zi * 0.0
            for i in range(len(zi)):

                if eazyp:
                    out = self.fit_combined_at_z(
                        z=zi[i], eazyp=eazyp, ix=ix, order=order, scale_fit=scale_fit
                    )
                    ci[i], bg, full, coeffs, err, scale_fit = out
                else:
                    out = self.fit_at_z(z=zi[i], templates=t_complex, fitter=fitter)
                    ci[i], bg, full, coeffs, err = out

                # out = self.fit_at_z(z=zi[i], templates=t_complex,
                #                     fitter=fitter)
                #
                # ci[i], bg, full, coeffs, err = out

                if verbose:
                    print("{0:.4f} - {1:10.1f}".format(zi[i], ci[i]))

            z = np.append(z, zi)
            chi2 = np.append(chi2, ci)

        so = np.argsort(z)
        z = z[so]
        chi2 = chi2[so]

        # Apply the prior
        if prior is not None:
            pz = np.interp(z, prior[0], prior[1])
            chi2 += pz

        # Get the fit with the individual line templates at the best redshift
        chi2x, bgz, fullz, coeffs, err = self.fit_at_z(
            z=z[np.argmin(chi2)], templates=t_i, fitter=fitter, get_uncertainties=True
        )

        # Mask outliers
        if outlier_threshold > 0:
            resid = self.scif - fullz - bgz
            outlier_mask = resid * self.sivarf < outlier_threshold
            # outlier_mask &= self.fit_mask
            # self.sivarf[outlier_mask] = 1/resid[outlier_mask]
            # print('Mask {0} pixels with resid > {1} sigma'.format((outlier_mask).sum(), outlier_threshold))

            print(
                "Mask {0} pixels with resid > {1} sigma".format(
                    (~outlier_mask & self.fit_mask).sum(), outlier_threshold
                )
            )
            self.fit_mask &= outlier_mask
            # self.DoF = self.fit_mask.sum() #(self.ivar > 0).sum()
            self.DoF = int((self.fit_mask * self.weightf).sum())

        # Table with z, chi-squared
        t = utils.GTable()
        t["z"] = z
        t["chi2"] = chi2

        if prior is not None:
            t["prior"] = pz

        # "area" parameter for redshift quality.
        num = np.trapz(np.clip(chi2 - chi2.min(), 0, 25), z)
        denom = np.trapz(z * 0 + 25, z)
        area25 = 1 - num / denom

        # "best" redshift
        zbest = z[np.argmin(chi2)]

        # Metadata will be stored as header keywords in the FITS table
        t.meta = OrderedDict()
        t.meta["ID"] = (self.h0["ID"], "Object ID")
        t.meta["RA"] = (self.h0["RA"], "Right Ascension")
        t.meta["DEC"] = (self.h0["DEC"], "Declination")
        t.meta["Z"] = (zbest, "Best-fit redshift")
        t.meta["CHIMIN"] = (chi2.min(), "Min Chi2")
        t.meta["CHIMAX"] = (chi2.max(), "Max Chi2")
        t.meta["DOF"] = (self.DoF, "Degrees of freedom")
        t.meta["AREA25"] = (area25, "Area under CHIMIN+25")
        t.meta["FITTER"] = (fitter, "Minimization algorithm")
        t.meta["HASPRIOR"] = (prior is not None, "Was prior specified?")

        # Best-fit templates
        tc, tl = self.generate_1D_templates(coeffs, templates_file=templates_file)
        # for i, te in enumerate(t_i):
        #     if i == 0:
        #         tc = t_i[te].zscale(0, scalar=coeffs[i])
        #         tl = t_i[te].zscale(0, scalar=coeffs[i])
        #     else:
        #         if te.startswith('line'):
        #             tc += t_i[te].zscale(0, scalar=0.)
        #         else:
        #             tc += t_i[te].zscale(0, scalar=coeffs[i])
        #
        #         tl += t_i[te].zscale(0, scalar=coeffs[i])

        # Get line fluxes, uncertainties and EWs
        il = 0
        for i, te in enumerate(t_i):
            if te.startswith("line"):
                il += 1

                if coeffs[i] == 0:
                    EW = 0.0
                else:
                    tn = t_i[te].zscale(0, scalar=coeffs[i]) + tc.zscale(0, scalar=1)

                    td = t_i[te].zscale(0, scalar=0) + tc.zscale(0, scalar=1)

                    clip = td.wave <= t_i[te].wave.max()
                    clip &= td.wave >= t_i[te].wave.min()

                    EW = np.trapz((tn.flux / td.flux)[clip] - 1, td.wave[clip])
                    if not np.isfinite(EW):
                        EW = -1000.0

                t.meta["LINE{0:03d}F".format(il)] = (
                    coeffs[i],
                    "{0} line flux".format(te[5:]),
                )

                t.meta["LINE{0:03d}E".format(il)] = (
                    err[i],
                    "{0} line flux uncertainty".format(te[5:]),
                )

                # print('xxx EW', EW)
                t.meta["LINE{0:03d}W".format(il)] = (
                    EW,
                    "{0} line rest EQW".format(te[5:]),
                )

        tfile = self.file.replace(".fits", ".zfit.fits")
        if os.path.exists(tfile):
            os.remove(tfile)

        t.write(tfile)

        # Add image HDU and templates
        hdu = pyfits.open(tfile, mode="update")
        hdu[1].header["EXTNAME"] = "ZFIT"

        # oned_templates = np.array([tc.wave*(1+zbest), tc.flux/(1+zbest),
        #                            tl.flux/(1+zbest)])
        header = pyfits.Header()
        header["TEMPFILE"] = (templates_file, "File with stored templates")
        hdu.append(pyfits.ImageHDU(data=coeffs, name="COEFFS"))

        for i in range(self.N):
            E = self.beams[i]
            model_i = fullz[self.slices[i]].reshape(E.sh)
            bg_i = bgz[self.slices[i]].reshape(E.sh)

            model_i[~np.isfinite(model_i)] = 0
            bg_i[~np.isfinite(bg_i)] = 0

            hdu.append(pyfits.ImageHDU(data=model_i, header=E.header, name="MODEL"))

            hdu.append(pyfits.ImageHDU(data=bg_i, header=E.header, name="BACKGROUND"))

        hdu.flush()

        if make_plot:
            self.make_fit_plot(hdu)

        return hdu

    @classmethod
    def generate_1D_templates(self, coeffs, templates_file="templates.npy"):
        """
        Generate 1D templates based on the given coefficients.
        
        Parameters
        ----------
        coeffs : array_like
            Coefficients used to scale the templates.
            
        templates_file : str
            Path to the file containing the templates. 
            Default is "templates.npy".

        Returns
        -------
        tc : ndarray
            The generated 1D template for the continuum component.

        tl : ndarray
            The generated 1D template for the line component.

        """

        t_complex, t_i = np.load(templates_file, allow_pickle=True)

        # Best-fit templates
        for i, te in enumerate(t_i):
            if i == 0:
                tc = t_i[te].zscale(0, scalar=coeffs[i])
                tl = t_i[te].zscale(0, scalar=coeffs[i])
            else:
                if te.startswith("line"):
                    tc += t_i[te].zscale(0, scalar=0.0)
                else:
                    tc += t_i[te].zscale(0, scalar=coeffs[i])

                tl += t_i[te].zscale(0, scalar=coeffs[i])

        return tc, tl

    def mask_drizzle_overlaps(self, threshold=3, verbose=True):
        """
        Mask pixels in beams of a given grism that are greater than
        `threshold` sigma times the minimum of all beams of that
        grism

        Parameters
        ----------

        threshold : float
            Threshold for masking. Default is 3.

        verbose : bool
            Print the number of pixels masked. Default is True.

        """
        min_grism = {}
        for grism in self.grisms:
            for E in self.beams:
                if not E.extver.startswith(grism):
                    continue

                if grism not in min_grism:
                    min_grism[grism] = E.scif * E.fit_mask
                else:
                    empty = (min_grism[grism] == 0) & E.fit_mask
                    min_grism[grism][E.fit_mask] = np.minimum(
                        min_grism[grism][E.fit_mask], E.scif[E.fit_mask]
                    )
                    min_grism[grism][empty] = E.scif[empty]

        for grism in self.grisms:
            for E in self.beams:
                if not E.extver.startswith(grism):
                    continue

                new_mask = (E.scif - min_grism[grism]) < threshold / E.sivarf
                if verbose:
                    print(
                        "Mask {0} additional pixels for ext {1}".format(
                            (~new_mask & E.fit_mask).sum(), E.extver
                        )
                    )

                E.fit_mask &= new_mask

    def make_fit_plot(self, hdu, scale_pz=True, show_2d=False):
        """
        Make a plot showing the fit

        Parameters
        ----------
        hdu : `~astropy.io.fits.HDUList`
            Fit results from `fit_zgrid`.

        scale_pz : bool
            Whether to scale the p(z) to DoF=1. Default is True.

        show_2d : bool
            Whether to show the 2D spectra. Default is False.

        Returns
        -------
        fig : `~matplotlib.figure.Figure`
            The figure object.

        """
        import matplotlib.pyplot as plt
        import matplotlib.gridspec
        from matplotlib.ticker import MultipleLocator

        # import grizli
        from . import utils

        zfit = utils.GTable.read(hdu[1])
        z = zfit["z"]
        chi2 = zfit["chi2"]

        # Initialize plot window
        Ng = len(self.grisms)
        if show_2d:
            height_ratios = [0.25] * self.N
            height_ratios.append(1)
            gs = matplotlib.gridspec.GridSpec(
                self.N + 1,
                2,
                width_ratios=[1, 1.5 + 0.5 * (Ng > 1)],
                height_ratios=height_ratios,
                hspace=0.0,
            )

            fig = plt.figure(figsize=[8 + 4 * (Ng > 1), 3.5 + 0.5 * self.N])
        else:
            gs = matplotlib.gridspec.GridSpec(
                1, 2, width_ratios=[1, 1.5 + 0.5 * (Ng > 1)], hspace=0.0
            )

            fig = plt.figure(figsize=[8 + 4 * (Ng > 1), 3.5])

        # Chi-squared
        axz = fig.add_subplot(gs[-1, 0])  # 121)

        axz.text(
            0.95,
            0.04,
            self.file + "\n" + "z={0:.4f}".format(z[np.argmin(chi2)]),
            ha="right",
            va="bottom",
            transform=axz.transAxes,
            fontsize=9,
        )

        # Scale p(z) to DoF=1.
        if scale_pz:
            scale_nu = chi2.min() / self.DoF
            scl_label = "_s"
        else:
            scale_nu = 1.0
            scl_label = ""

        axz.plot(z, (chi2 - chi2.min()) / scale_nu, color="k")
        axz.fill_between(z, (chi2 - chi2.min()) / scale_nu, 27, color="k", alpha=0.5)
        axz.set_ylim(-4, 27)
        axz.set_xlabel(r"$z$")
        axz.set_ylabel(
            r"$\Delta\chi^2{2}$ ({0:.0f}/$\nu$={1:d})".format(
                chi2.min(), self.DoF, scl_label
            )
        )
        axz.set_yticks([1, 4, 9, 16, 25])

        axz.set_xlim(z.min(), z.max())
        axz.grid()

        # 2D spectra
        if show_2d:
            twod_axes = []
            for i in range(self.N):
                ax_i = fig.add_subplot(gs[i, 1])

                model = hdu["MODEL", self.ext[i]].data
                ymax = model[np.isfinite(model)].max()
                # print('MAX', ymax)

                cmap = "viridis_r"
                cmap = "cubehelix_r"

                clean = self.beams[i].sci - hdu["BACKGROUND", self.ext[i]].data
                clean *= self.beams[i].fit_mask.reshape(self.beams[i].sh)

                w = self.beams[i].wave / 1.0e4

                ax_i.imshow(
                    clean,
                    vmin=-0.02 * ymax,
                    vmax=1.1 * ymax,
                    origin="lower",
                    extent=[w[0], w[-1], 0.0, 1.0],
                    aspect="auto",
                    cmap=cmap,
                )

                ax_i.text(
                    0.04,
                    0.92,
                    self.ext[i],
                    ha="left",
                    va="top",
                    transform=ax_i.transAxes,
                    fontsize=8,
                )

                ax_i.set_xticklabels([])
                ax_i.set_yticklabels([])
                twod_axes.append(ax_i)

        axc = fig.add_subplot(gs[-1, 1])  # 224)

        # 1D Spectra + fit
        ymin = 1.0e30
        ymax = -1.0e30
        wmin = 1.0e30
        wmax = -1.0e30

        for i in range(self.N):

            E = self.beams[i]

            clean = E.sci - hdu["BACKGROUND", self.ext[i]].data
            w, fl, er = E.optimal_extract(clean)
            w, flm, erm = E.optimal_extract(hdu["MODEL", self.ext[i]].data)
            w = w / 1.0e4

            # Do we need to convert to F-lambda units?
            if E.is_flambda:
                unit_corr = 1.0
                clip = (er > 0) & np.isfinite(er) & np.isfinite(flm)
                clip[:10] = False
                clip[-10:] = False

                if clip.sum() == 0:
                    clip = er > -1

            else:
                unit_corr = 1.0 / E.sens
                clip = E.sens > 0.1 * E.sens.max()
                clip &= (np.isfinite(flm)) & (er > 0)

            if clip.sum() == 0:
                continue

            # fl *= 100*unit_corr
            # er *= 100*unit_corr
            # flm *= 100*unit_corr

            fl *= unit_corr / 1.0e-19
            er *= unit_corr / 1.0e-19
            flm *= unit_corr / 1.0e-19

            f_alpha = (
                1.0 / self.Ngrism[E.grism] ** 0.5
            )  # self.h0['N{0}'.format(E.header['GRISM'])]**0.5

            axc.errorbar(
                w[clip],
                fl[clip],
                er[clip],
                color=GRISM_COLORS[E.grism],
                alpha=0.3 * f_alpha,
                marker=".",
                linestyle="None",
            )
            # axc.fill_between(w[clip], (fl+er)[clip], (fl-er)[clip], color='k', alpha=0.2)
            axc.plot(w[clip], flm[clip], color="r", alpha=0.6 * f_alpha, linewidth=2)

            # Plot limits
            # ymax = np.maximum(ymax, (flm+np.median(er))[clip].max())
            # ymin = np.minimum(ymin, (flm-er*0.)[clip].min())

            ymax = np.maximum(ymax, np.percentile((flm + np.median(er))[clip], 98))

            ymin = np.minimum(ymin, np.percentile((flm - er * 0.0)[clip], 2))

            wmax = np.maximum(wmax, w.max())
            wmin = np.minimum(wmin, w.min())

        axc.set_xlim(wmin, wmax)
        axc.semilogx(subsx=[wmax])
        # axc.set_xticklabels([])
        axc.set_xlabel(r"$\lambda$")
        axc.set_ylabel(r"$f_\lambda \times 10^{-19}$")
        # axc.xaxis.set_major_locator(MultipleLocator(0.1))

        axc.set_ylim(ymin - 0.2 * ymax, 1.2 * ymax)
        axc.grid()

        for ax in [axc]:  # [axa, axb, axc]:

            labels = np.arange(np.ceil(wmin * 10), np.ceil(wmax * 10)) / 10.0
            ax.set_xticks(labels)
            ax.set_xticklabels(labels)
            # ax.set_xticklabels([])
            # print(labels, wmin, wmax)

        if show_2d:
            for ax in twod_axes:
                ax.set_xlim(wmin, wmax)

        gs.tight_layout(fig, pad=0.1, w_pad=0.1)
        fig.savefig(self.file.replace(".fits", ".zfit.png"))
        return fig


class StackedSpectrum(object):
    def __init__(
        self,
        file="gnt_18197.stack.G141.285.fits",
        sys_err=0.02,
        mask_min=0.1,
        extver="G141",
        mask_threshold=7,
        fcontam=1.0,
        min_ivar=0.001,
        MW_EBV=0.0,
    ):
        """
        The StackedSpectrum class.

        Parameters
        ----------
        file : str
            The path to the FITS file containing the data. 
            Default is "gnt_18197.stack.G141.285.fits".

        sys_err : float
            The systematic error to be used in the calculations. 
            Default is 0.02.

        mask_min : float
            The minimum value for the mask. Default is 0.1.

        extver : str
            The extension version. Default is "G141".

        mask_threshold : int
            The threshold value for the mask. Default is 7.

        fcontam : float
            The contamination weighting factor. Default is 1.0.

        min_ivar : float
            The minimum inverse variance value. Default is 0.001.

        MW_EBV : float
            The Milky Way extinction value. Default is 0.0.

        """
        
        import os

        # import grizli
        from . import GRIZLI_PATH
        from . import grismconf

        self.sys_err = sys_err
        self.mask_min = mask_min
        self.extver = extver
        self.grism = self.extver.split(",")[0]
        self.mask_threshold = mask_threshold

        self.MW_EBV = MW_EBV
        self.init_galactic_extinction(MW_EBV)

        self.file = file
        self.hdulist = pyfits.open(file)

        self.h0 = self.hdulist[0].header.copy()
        self.header = self.hdulist["SCI", extver].header.copy()
        self.sh = (self.header["NAXIS2"], self.header["NAXIS1"])
        self.wave = self.get_wavelength_from_header(self.header)
        self.wavef = np.dot(np.ones((self.sh[0], 1)), self.wave[None, :]).flatten()

        if "BEAM" in self.header:
            self.beam_name = self.header["BEAM"]
        else:
            self.beam_name = "A"

        # Configuration file
        self.is_flambda = self.header["ISFLAM"]
        self.conf_file = self.header["CONF"]
        try:
            self.conf = grismconf.aXeConf(self.conf_file)
        except:
            # Try global path
            base = os.path.basename(self.conf_file)
            localfile = os.path.join(GRIZLI_PATH, "CONF", base)
            self.conf = grismconf.aXeConf(localfile)

        self.conf.get_beams()

        self.sci = self.hdulist["SCI", extver].data * 1.0
        self.ivar0 = self.hdulist["WHT", extver].data * 1
        self.size = self.sci.size
        self.thumbs = {}

        self.scif = self.sci.flatten()
        self.ivarf0 = self.ivar0.flatten()

        self.ivarf = 1 / (1 / self.ivarf0 + (sys_err * self.scif) ** 2)
        self.ivar = self.ivarf.reshape(self.sh)

        self.sivarf = np.sqrt(self.ivar).flatten()

        self.fit_mask = self.ivarf > min_ivar * self.ivarf.max()
        self.fit_mask &= np.isfinite(self.scif) & np.isfinite(self.ivarf)

        # Dummy parameter.  Implemented for MultiBeam
        self.sens_mask = 1.0

        # Contamination weighting
        self.fcontam = fcontam
        if ("CONTAM", extver) in self.hdulist:
            self.contam = np.abs(self.hdulist["CONTAM", extver].data * 1.0)
            self.weight = np.exp(-fcontam * self.contam * np.sqrt(self.ivar0))
            self.contamf = self.contam.flatten()
            self.weightf = self.weight.flatten()
        else:
            self.contam = self.sci * 0.0
            self.contamf = self.scif * 0.0

            self.weight = self.sci * 0.0 + 1
            self.weightf = self.scif * 0.0 + 1

        # Spatial kernel
        self.kernel = self.hdulist["KERNEL", extver].data * 1
        self.kernel /= self.kernel.sum()

        self._build_model()

        self.flat = self.compute_model()

        if self.is_flambda:
            elec = self.flat * self.ivarf0 / self.ivarf0.max()
        else:
            elec = self.flat

        self.fit_mask &= elec > mask_min * elec.max()

        self.DoF = self.fit_mask.sum()  # (self.ivar > 0).sum()

        if mask_threshold > 0:
            self.drizzle_mask(mask_threshold=mask_threshold)

        self.flat_flam = self.compute_model(in_place=False, is_cgs=True)

    def init_galactic_extinction(self, MW_EBV=0.0, R_V=utils.MW_RV):
        """
        Initialize Fitzpatrick 99 Galactic extinction

        Parameters
        ----------
        MW_EBV : float
            Local E(B-V)

        R_V : float
            Relation between specific and total extinction,
            ``a_v = r_v * ebv``.

        Returns
        -------
        Sets `self.MW_F99` attribute, which is a callable function that
        returns the extinction for a supplied array of wavelengths.

        If MW_EBV <= 0, then sets `self.MW_F99 = None`.

        """
        self.MW_F99 = None
        if MW_EBV > 0:
            self.MW_F99 = utils.MW_F99(MW_EBV * R_V, r_v=R_V)

            # try:
            #     from specutils.extinction import ExtinctionF99
            #     self.MW_F99 = ExtinctionF99(MW_EBV*R_V, r_v=R_V)
            # except(ImportError):
            #     print('Couldn\'t import `specutils.extinction`, MW extinction not implemented')

    @classmethod
    def get_wavelength_from_header(self, h):
        """
        Generate wavelength array from WCS keywords

        Parameters
        ----------
        h : `astropy.io.fits.Header`
            The FITS header containing the WCS keywords.

        Returns
        -------
        w : `~np.array`
            The generated wavelength array.

        """
        w = (np.arange(h["NAXIS1"]) + 1 - h["CRPIX1"]) * h["CD1_1"] + h["CRVAL1"]

        # Now header keywords scaled to microns
        if w.max() < 3:
            w *= 1.0e4

        return w

    def init_optimal_profile(self):
        """Initilize optimal extraction profile"""
        m = self.compute_model(in_place=False)
        m = m.reshape(self.sh)
        m[m < 0] = 0
        self.optimal_profile = m / m.sum(axis=0)

    def optimal_extract(self, data, bin=0, ivar=None, **kwargs):
        """
        Optimally-weighted 1D extraction

        Parameters
        ----------
        data : `~np.ndarray`
            2D array of the data to extract.

        bin : int
            Binning factor. Default is 0.

        ivar : array-like
            Inverse variance to replace `self.ivar`

        Returns
        -------
        wave : `~np.array`
            Wavelength array
        
        opt_flux : `~np.array`
            Optimal flux array

        opt_rms : `~np.array`
            Optimal RMS array

        """
        import scipy.ndimage as nd

        if not hasattr(self, "optimal_profile"):
            self.init_optimal_profile()

        prof = self.optimal_profile

        if ivar is None:
            ivar = self.ivar

        num = prof * data * ivar * self.weight
        den = prof ** 2 * ivar * self.weight
        opt_flux = num.sum(axis=0) / den.sum(axis=0)
        opt_var = 1.0 / den.sum(axis=0)

        opt_rms = np.sqrt(opt_var)
        clip = (opt_var == 0) | ~np.isfinite(opt_var)
        opt_rms[clip] = 0
        opt_flux[clip] = 0

        if bin > 1:
            kern = np.ones(bin, dtype=float) / bin
            opt_flux = nd.convolve(opt_flux, kern)[bin // 2 :: bin]
            opt_var = nd.convolve(opt_var, kern ** 2)[bin // 2 :: bin]
            opt_rms = np.sqrt(opt_var)
            wave = self.wave[bin // 2 :: bin]
        else:
            wave = self.wave

        return wave, opt_flux, opt_rms

    def _build_model(self):
        """
        Initiazize components for generating 2D model
        """
        from .utils_numba.interp import interp_conserve_c

        NY = self.sh[0]
        data = np.zeros(
            (self.header["NAXIS1"], self.header["NAXIS2"], self.header["NAXIS1"])
        )

        for j in range(NY // 2):
            data[j, :, : j + NY // 2] += self.kernel[:, -NY // 2 - j :]

        for j in range(self.sh[1] - NY // 2, self.sh[1]):
            data[j, :, -NY // 2 + j :] += self.kernel[:, : self.sh[1] - j + NY // 2]

        for j in range(NY // 2, self.sh[1] - NY // 2):
            data[j, :, j - NY // 2 : j + NY // 2] += self.kernel

        self.fit_data = data.reshape(self.sh[1], -1)

        if not self.is_flambda:
            conf_sens = self.conf.sens[self.beam_name]
            if self.MW_F99 is not None:
                MWext = 10 ** (-0.4 * (self.MW_F99(conf_sens["WAVELENGTH"] * u.AA)))
            else:
                MWext = 1.0

            sens = interp_conserve_c(
                self.wave, conf_sens["WAVELENGTH"], conf_sens["SENSITIVITY"] * MWext
            )

            if "DLAM0" in self.header:
                # print('xx Header')
                dlam = self.header["DLAM0"]
            else:
                dlam = np.median(np.diff(self.wave))

            # dlam = np.median(np.diff(self.wave))
            dlam = np.gradient(self.wave)

            # G800 variable dispersion, which isn't coming out right
            # in the 2D drizzled spectrum
            if self.grism == "G800L":
                # Linear fit to differential G800L dispersion
                # y = np.diff(beam.wave)/np.diff(beam.wave)[0]
                # x = beam.wave[1:]
                # c_i = np.polynomial.Polynomial.fit(x/1.e4, y, deg=1)
                c_i = np.array([0.98928126, 0.07498747])
                scale = np.polynomial.Polynomial(c_i)(self.wave / 1.0e4)
                sens *= scale

            self.sens = sens * dlam  # *1.e-17
            self.fit_data = (self.fit_data.T * self.sens).T

    def compute_model(self, spectrum_1d=None, is_cgs=None, in_place=False):
        """
        Generate the model spectrum

        Parameters
        ----------
        spectrum_1d : tuple
            A tuple containing the wavelength and flux arrays. Default is None.

        is_cgsm, in_place : dummy parameters to match `MultiBeam.compute_model`.

        Returns
        -------
        model : `~np.ndarray`
            The generated model spectrum.

        """
        from .utils_numba.interp import interp_conserve_c

        if spectrum_1d is None:
            fl = self.wave * 0 + 1
        else:
            fl = interp_conserve_c(self.wave, spectrum_1d[0], spectrum_1d[1])

        model = np.dot(fl, self.fit_data)  # .reshape(self.sh)
        return model



"""
Helper function for running Grizli redshift fits in AWS lambda

event = {'s3_object_path' : 'Pipeline/j001452+091221/Extractions/j001452+091221_00277.beams.fits'}

Optional event key (and anything to grizli.fitting.run_all):

    'skip_started' : Look for a start.log file and abort if found

    'check_wcs' : check for WCS files, needed for ACS fits

    'quasar_fit' : run fit with quasar templates

    'output_path' : optional output path in the aws-grivam bucket

Example run_all arguments:

    'use_psf' : Use point source models (for quasar fits)

    'verbose' : verbose output

    'zr' : [zmin, zmax] Redshift fitting range

"""

import os
import glob
import time

import numpy as np

FALSE_OPTIONS = [False, 'false', 'False', 0, '0', 'n', 'N']
TRUE_OPTIONS = [True,  'true',  'True',  1, '1', 'y', 'Y']


def check_object_in_footprint(id, wcs_fits, cat, rd=None):
    """
    Check that a given object is within a WCS footprint

    Parameters
    ----------
    id : int
        Integer id of the catalog objects

    wcs_fits : str
        WCS filename, like 'ibfuw1psq_flt.01.wcs.fits'.

    cat : `~astropy.table.Table`
        Table object of the "ir_cat.fits" source detection table, with
        SExtractor columns NUMBER, X_WORLD, Y_WORLD.

    """
    import matplotlib.path
    import astropy.wcs as pywcs
    import astropy.io.fits as pyfits

    if rd is None:
        ix = cat['NUMBER'] == id
        ra, dec = cat['X_WORLD'][ix][0], cat['Y_WORLD'][ix][0]
    else:
        ra, dec = rd

    im = pyfits.open(wcs_fits)
    im[0].header['NAXIS'] = 2
    im[0].header['NAXIS1'] = im[0].header['CRPIX1']*2
    im[0].header['NAXIS2'] = im[0].header['CRPIX2']*2

    wcs = pywcs.WCS(im[0].header, fobj=im, relax=True)
    fp = matplotlib.path.Path(wcs.calc_footprint())
    has_point = fp.contains_point([ra, dec])
    im.close()

    return has_point


def extract_beams_from_flt(root, bucket, id, size=32, clean=True, silent=False):
    """
    Download GrismFLT files and extract the beams file
    """
    import gc
    import boto3

    import matplotlib.pyplot as plt

    import grizli
    from grizli import fitting, utils, multifit
    #from grizli import __version__ as grizli__version
    grizli__version = grizli.__version__
    
    utils.set_warnings()
    from grizli.pipeline import auto_script

    s3 = boto3.resource('s3')
    s3_client = boto3.client('s3')
    bkt = s3.Bucket(bucket)

    # WCS files for ACS
    files = [obj.key for obj in 
             bkt.objects.filter(Prefix=f'HST/Pipeline/{root}/Extractions/j')]
    files += [obj.key for obj in 
              bkt.objects.filter(Prefix=f'HST/Pipeline/{root}/Extractions/i')]

    prefix = f'HST/Pipeline/{root}/Extractions/{root}-ir.cat.fits'
    files += [obj.key for obj in bkt.objects.filter(Prefix=prefix)]

    prefix = f'HST/Pipeline/{root}/Extractions/fit_args.npy'
    files += [obj.key for obj in bkt.objects.filter(Prefix=prefix)]

    download_files = []
    for file in np.unique(files):
        if ('cat.fits' in file) | ('fit_args' in file):
            if os.path.exists(os.path.basename(file)):
                continue

            download_files.append(file)

    for file in download_files:
        print(file)
        bkt.download_file(file, os.path.basename(file),
                          ExtraArgs={"RequestPayer": "requester"})

    # Read the catalog
    ircat = utils.read_catalog('{0}-ir.cat.fits'.format(root))
    ix = ircat['NUMBER'] == id
    object_rd = (ircat['X_WORLD'][ix], ircat['Y_WORLD'][ix])
    del(ircat)

    # One beam at a time
    beams = None

    flt_files = []
    for file in files:
        if 'GrismFLT.fits' in file:
            flt_files.append(file)

    if not silent:
        print('Read {0} GrismFLT files'.format(len(flt_files)))
    
    if os.path.exists('{0}_fit_args.npy'.format(root)):
        args_file = '{0}_fit_args.npy'.format(root)
    else:
        args_file = 'fit_args.npy'

    for i, file in enumerate(flt_files):
        if not silent:
            print('# Read {0}/{1}'.format(i+1, len(flt_files)))

        flt, ext, _, _ = os.path.basename(file).split('.')
        if flt.startswith('i'):
            fl = 'flt'
        elif flt.startswith('jw'):
            fl = 'rate'
        else:
            fl = 'flc'

        out_files = ['{0}_{2}.{1}.wcs.fits'.format(flt, ext, fl),
                     '{0}.{1}.GrismFLT.fits'.format(flt, ext),
                     '{0}.{1}.GrismFLT.pkl'.format(flt, ext)]

        exp_has_id = False

        for j, f_j in enumerate(out_files):
            aws_file = os.path.join(os.path.dirname(file), f_j)
            if not silent:
                print('  ', aws_file)

            if not os.path.exists(f_j):
                bkt.download_file(aws_file, f_j,
                                  ExtraArgs={"RequestPayer": "requester"})

            # WCS file, check if object in footprint
            if f_j.endswith('.wcs.fits'):
                exp_has_id = check_object_in_footprint(None, f_j, None,
                                                       rd=object_rd)
                if not exp_has_id:
                    if clean:
                        os.remove(f_j)
                    break

        if not exp_has_id:
            continue

        beams_i = auto_script.extract(field_root=root, maglim=[13, 24],
                                      prior=None, MW_EBV=0.00, ids=id,
                                      pline={}, fit_only_beams=True,
                                      run_fit=False, poly_order=7,
                                      master_files=[os.path.basename(file)],
                                      grp=None, bad_pa_threshold=None,
                                      fit_trace_shift=False, size=size,
                                      diff=True, min_sens=0.02,
                                      skip_complete=True, fit_args={},
                                      args_file=args_file,
                                      get_only_beams=True)

        # Remove the GrismFLT file
        for f_j in out_files:
            if ('GrismFLT' in f_j) & clean:
                os.remove(f_j)

        if beams is None:
            beams = beams_i
        else:
            beams.extend(beams_i)

    # Garbage collector
    gc.collect()

    if not beams:
        print('No beams found for {0} id={1}'.format(root, id))
        return False

    # Grism Object
    args = np.load(args_file, allow_pickle=True)[0]
    if '_ehn_' in root:
        try:
            import wfc3dash.grism.grism
            mb = wfc3dash.grism.grism.run_align_dash(beams, args)
        except ImportError:
            mb = multifit.MultiBeam(beams, **args)
    else:
        mb = multifit.MultiBeam(beams, **args)
        
    mb.write_master_fits()

    # 1D spectrum with R=30 fit
    if True:
        bin_steps, step_templ = utils.step_templates(wlim=[5000, 18000.0],
                                                     R=30, round=10)

        tfit = mb.template_at_z(z=0, templates=step_templ,
                                fit_background=True, fitter='lstsq',
                                get_uncertainties=2)

        fig1 = mb.oned_figure(figsize=[5, 3], tfit=tfit, show_beams=True,
                              scale_on_stacked=True, ylim_percentile=5)

        outroot = '{0}_{1:05d}.R{2:.0f}'.format(root, id, 30)
        hdu = mb.oned_spectrum_to_hdu(outputfile=outroot+'.fits',
                                              tfit=tfit, wave=bin_steps)

        fig1.savefig(outroot+'.png')
        del(hdu)

        # Drizzled spectrum
        hdu, fig = mb.drizzle_grisms_and_PAs(fcontam=args['fcontam'],
                                             flambda=False,
                                             kernel='point', size=32,
                                             tfit=tfit, diff=False)

        hdu[0].header['GRIZLIV'] = (grizli__version, 'Grizli version')

        fig.savefig('{0}_{1:05d}.stack.png'.format(root, id))

        hdu.writeto('{0}_{1:05d}.stack.fits'.format(root, id),
                    overwrite=True)

        plt.close('all')
        del(hdu)

    outfiles = ['{0}_{1:05d}.beams.fits'.format(root, id)]
    outfiles += glob.glob(outroot+'*')
    outfiles += glob.glob('{0}_{1:05d}.stack*'.format(root, id))

    return(outfiles)


def run_grizli_fit(event):
    import boto3
    import json
    import shutil
    import gc

    import matplotlib.pyplot as plt

    import grizli
    from grizli import fitting, utils, multifit

    try:
        from grizli.aws import db as grizli_db
        dbFLAGS = grizli_db.FLAGS
    except:
        pass

    utils.set_warnings()

    #event = {'s3_object_path':'HST/Pipeline/j001452+091221/Extractions/j001452+091221_00277.beams.fits'}

    silent = False
    if 'silent' in event:
        silent = event['silent'] in TRUE_OPTIONS

    update_db = True
    if 'update_db' in event:
        update_db = event['update_db'] in TRUE_OPTIONS
    
    ###
    # Parse event arguments
    ###
    event_kwargs = {}
    for k in event:

        # Lists
        if isinstance(event[k], str):
            # Split lists
            if ',' in event[k]:
                try:
                    event_kwargs[k] = np.asarray(event[k].split(','),dtype=float)
                except:
                    event_kwargs[k] = event[k].split(',')
            else:
                event_kwargs[k] = event[k]
        else:
            try:
                event_kwargs[k] = json.loads(event[k])
            except:
                event_kwargs[k] = event[k]
    
    # Defaults
    if 'skip_started' not in event_kwargs:
        event_kwargs['skip_started'] = True
    
    if 'size' not in event_kwargs:
        event_kwargs['size'] = 32
        
    for k in ['quasar_fit', 'extract_from_flt', 'fit_stars', 'beam_info_only']:
        if k not in event_kwargs:
            event_kwargs[k] = False

    if event_kwargs['beam_info_only'] in TRUE_OPTIONS:
        dbtable = 'multibeam_v2'
    elif event_kwargs['quasar_fit'] in TRUE_OPTIONS:
        dbtable = 'redshift_fit_v2_quasar'
    elif event_kwargs['fit_stars'] in TRUE_OPTIONS:
        dbtable = 'stellar_fit_v2'
    else:
        dbtable = 'redshift_fit_v2'
    
    if 'dbtable' in event_kwargs:
        dbtable = event_kwargs['dbtable']
        
    if not silent:
        print('Grizli version: ', grizli.__version__)

    # Disk space
    total, used, free = shutil.disk_usage("/")
    if not silent:
        print('Disk info: Total = {0:.2f} / Used = {1:.2f} / Free = {2:.2f}'.format(total // (2**20), used // (2**20), free // (2**20)))

    # Output path
    if 'output_path' in event:
        output_path = event['output_path']
    else:
        output_path = None

    if 'bucket' in event:
        event_kwargs['bucket'] = event['bucket']
    else:
        event_kwargs['bucket'] = 'grizli-v2' #'aws-grivam'

    if 'working_directory' in event:
        os.chdir(event['working_directory'])
    else:
        os.chdir('/tmp/')

    if not silent:
        print('Working directory: {0}'.format(os.getcwd()))

    files = glob.glob('*')
    files.sort()

    # Filenames, etc.
    beams_file = os.path.basename(event['s3_object_path'])
    root = '_'.join(beams_file.split('_')[:-1])
    id = int(beams_file.split('_')[-1].split('.')[0])

    try:
        db_status = grizli_db.get_redshift_fit_status(root, id, table=dbtable)
    except:
        db_status = -1
        update_db = False
        
    # Initial log
    start_log = '{0}_{1:05d}.start.log'.format(root, id)
    full_start = 'HST/Pipeline/{0}/Extractions/{1}'.format(root, start_log)
    if ((start_log in files) | (db_status >= 0)) & event_kwargs['skip_started']:
        print('Log file {0} found in {1} (db_status={2})'.format(start_log, os.getcwd(), db_status))
        return True

    if not silent:
        for i, file in enumerate(files):
            print('Initial file ({0}): {1}'.format(i+1, file))

    if os.path.exists('{0}/matplotlibrc'.format(grizli.GRIZLI_PATH)):
        os.system('cp {0}/matplotlibrc .'.format(grizli.GRIZLI_PATH))

    s3 = boto3.resource('s3')
    s3_client = boto3.client('s3')
    bkt = s3.Bucket(event_kwargs['bucket'])
    
    ubkt = bkt
    if 'up_bucket' in event_kwargs:
        if event_kwargs['up_bucket'] != event_kwargs['bucket']:
            print('Set upload bucket: ' + event_kwargs['up_bucket'])
            ubkt = s3.Bucket(event_kwargs['up_bucket'])
            
    if event_kwargs['skip_started']:
        res = [r.key for r in ubkt.objects.filter(Prefix=full_start)]
        if res:
            print(f'Already started ({start_log}), aborting.')
            return True

    fp = open(start_log, 'w')
    fp.write(time.ctime()+'\n')
    fp.close()
    if update_db:
        ubkt.upload_file(start_log, full_start)

    # Download fit arguments
    if 'force_args' in event:
        force_args = event['force_args'] in TRUE_OPTIONS
    else:
        force_args = False

    args_files = [f'{root}_fit_args.npy', 'fit_args.npy']
    for args_file in args_files:
        if (not os.path.exists(args_file)) | force_args:
            aws_file = f'HST/Pipeline/{root}/Extractions/{args_file}'
            try:
                bkt.download_file(aws_file, f'./{args_file}',
                              ExtraArgs={"RequestPayer": "requester"})
                print(f'Use args_file = {args_file}')
                break
            except:
                continue
        else:
            print(f'Use found args_file = {args_file}')
            break
    
    # If no beams file in the bucket, try to generate it
    put_beams = False
    try:
        if not os.path.exists(beams_file):
            bkt.download_file(event['s3_object_path'], 
                              f'./{beams_file}',
                              ExtraArgs={"RequestPayer": "requester"})
            put_beams = False
    except:
        print('Extract from GrismFLT object!')
        if 'clean' in event:
            if isinstance(event['clean'], str):
                run_clean = event['clean'].lower() in ['true', 'y', '1']
            else:
                run_clean = event['clean']
        else:
            run_clean = True

        if update_db:
            try:
                # Extracting beams
                grizli_db.update_redshift_fit_status(root, id,
                                                status=dbFLAGS['start_beams'],
                                                    table=dbtable)
            except:
                print('Set DB flag failed: start_beams')
                pass

        status = extract_beams_from_flt(root, event_kwargs['bucket'], id,
                                        size=event_kwargs['size'],
                                        clean=run_clean, silent=silent)

        # Garbage collector
        gc.collect()

        if status is False:
            return False
        else:
            beams_file = status[0]
        
        if update_db:
            try:
                # Beams are done
                grizli_db.update_redshift_fit_status(root, id,
                                                status=dbFLAGS['done_beams'],
                                                     table=dbtable)
            except:
                pass

        put_beams = True

        # upload it now
        output_path = f'HST/Pipeline/{root}/Extractions'
        if update_db:
            for outfile in status:
                aws_file = f'{output_path}/{outfile}'
                print(aws_file)
                bkt.upload_file(outfile, aws_file, 
                            ExtraArgs={'ACL': 'public-read'})

    if ('run_fit' in event) & (dbtable == 'redshift_fit_v2') & update_db:
        if event['run_fit'] in FALSE_OPTIONS:
            res = ubkt.delete_objects(Delete={'Objects': 
                                               [{'Key': full_start}]})

            try:
                grizli_db.update_redshift_fit_status(root, id,
                                                 status=dbFLAGS['no_run_fit'],
                                                 table=dbtable)
            except:
                pass

            return True

    utils.fetch_acs_wcs_files(beams_file, bucket_name=event_kwargs['bucket'])

    # Update the multibeam/beam_geometry tables
    if os.path.exists(beams_file) & update_db:
        args = np.load(args_file, allow_pickle=True)[0]
        for arg in event_kwargs:
            if arg in args:
                args[arg] = event_kwargs[arg]

        grizli_db.multibeam_to_database(beams_file, Rspline=15, force=False,
                                        **args)

    if (dbtable == 'multibeam_v2') & update_db:
        # Done
        res = ubkt.delete_objects(Delete={'Objects': [{'Key': full_start}]})
        return True

    # Download WCS files
    # if event_kwargs['check_wcs']:
    #     # WCS files for ACS
    #     files = [obj.key for obj in bkt.objects.filter(Prefix='Pipeline/{0}/Extractions/j'.format(root))]
    #     for file in files:
    #         if 'wcs.fits' in file:
    #             if os.path.exists(os.path.basename(file)):
    #                 continue
    #
    #             bkt.download_file(file, os.path.basename(file),
    #                               ExtraArgs={"RequestPayer": "requester"})

    # Is zr in the event dict?
    # if 'zr' in event:
    #     zr = list(np.asarray(event['zr'],dtype=float))
    # else:
    #     try:
    #         zr = np.load('fit_args.npy')[0]['zr']
    #     except:
    #         zr = np.load('fit_args.npy', allow_pickle=True)[0]['zr']

    # Directory listing
    files = glob.glob('*')
    files.sort()

    if not silent:
        for i, file in enumerate(files):
            print(f'File ({i+1}): {file}')

    try:
        files = glob.glob(f'{root}_{id:05d}*R30.fits')
        if (len(files) > 0) & (dbtable == 'redshift_fit_v2'):
            grizli_db.send_1D_to_database(files=files)
    except:
        print('Failed to send R30 to spec1d database')
        pass

    ###
    # Run the fit
    if update_db:
        try:
            grizli_db.update_redshift_fit_status(root, id, table=dbtable,
                                        status=dbFLAGS['start_redshift_fit'])
        except:
            print('Set DB flag failed: start_redshift_fit')
            pass
    
    # Extra args from numpy file on s3
    if 's3_argsfile' in event_kwargs:
        if os.path.exists(event_kwargs['s3_argsfile']):
            result = (event_kwargs['s3_argsfile'], 1)
        else:
            result = utils.fetch_s3_url(url=event_kwargs['s3_argsfile'],
                                    file_func=lambda x : os.path.join('./',x), 
                                    skip_existing=True, verbose=True)
        
        s3_argsfile_local, status = result
        if os.path.exists(s3_argsfile_local):
            try:
                extra_args = np.load(s3_argsfile_local, allow_pickle=True)[0]
                
                for k in extra_args:
                    print(f'{s3_argsfile_local} {k}')
                    event_kwargs[k] = extra_args[k]
                
            except OSError:
                print(f'Failed to get keywords from {s3_argsfile_local}')
                
    if event_kwargs['quasar_fit'] in TRUE_OPTIONS:

        # Don't recopy beams file
        put_beams = False

        # Don't make line maps
        if 'min_line_sn' not in event_kwargs:
            event_kwargs['min_line_sn'] = np.inf

        # Don't make drizzled psfs
        if 'get_ir_psfs' not in event_kwargs:
            event_kwargs['get_ir_psfs'] = False

        # Fit line widths
        if 'get_line_width' not in event_kwargs:
            event_kwargs['get_line_width'] = True

        # sys_err
        if 'sys_err' not in event_kwargs:
            event_kwargs['sys_err'] = 0.05

        # Don't use photometry
        event_kwargs['phot_obj'] = None
        event_kwargs['use_phot_obj'] = False

        event_kwargs['fit_only_beams'] = True
        event_kwargs['fit_beams'] = False

        templ_args = {'uv_line_complex': True,
                      'broad_fwhm': 2800,
                      'narrow_fwhm': 1000,
                      'fixed_narrow_lines': True,
                      'Rspline': 15,
                      'include_reddened_balmer_lines': False}

        for k in templ_args:
            if k in event_kwargs:
                templ_args[k] = event_kwargs.pop(k)

        if templ_args['broad_fwhm'] < 0:
            use_simple_templates = True
            templ_args['broad_fwhm'] *= -1
        else:
            use_simple_templates = False

        print('load_quasar_templates(**{0})'.format(templ_args))
        q0, q1 = utils.load_quasar_templates(**templ_args)

        if use_simple_templates:
            x0 = utils.load_templates(full_line_list=['highO32'], 
                                      continuum_list=['quasar_lines.txt', 
                                          'red_blue_continuum_noLya.txt'],
                                      line_complexes=False, fwhm=1000)

            x0 = utils.load_templates(full_line_list=[], 
                                      continuum_list=['quasar_lines.txt',
                                            'red_blue_continuum_noLya.txt',
                                            'fsps_starburst_lines.txt'], 
                                      line_complexes=False, fwhm=1000)

            # for t in q0:
            #     if 'bspl' in t:
            #         x0[t] = q0[t]

            # Rename templates
            x0['continuum spline'] = x0.pop('red_blue_continuum_noLya.txt')
            x0['quasar'] = x0.pop('quasar_lines.txt')
            x0['fsps_starburst'] = x0.pop('fsps_starburst_lines.txt')

            q0 = x0
            q1['continuum spline'] = x0['continuum spline']

        # Quasar templates with fixed line ratios
        # q0, q1 = utils.load_quasar_templates(uv_line_complex=True,
        #                                     broad_fwhm=2800, narrow_fwhm=1000,
        #                                     fixed_narrow_lines=True,
        #                                     Rspline=15)

        if 'zr' not in event_kwargs:
            event_kwargs['zr'] = [0.03, 6.8]
        if 'fitter' not in event_kwargs:
            event_kwargs['fitter'] = ['lstsq', 'lstsq']

        print('run_all_parallel: {0}'.format(event_kwargs))

        fitting.run_all_parallel(id, t0=q0, t1=q1, args_file=args_file,
                                 **event_kwargs)

        if output_path is None:
            #output_path = 'Pipeline/QuasarFit'.format(root)
            output_path = 'HST/Pipeline/{0}/Extractions'.format(root)

    elif event_kwargs['fit_stars'] in TRUE_OPTIONS:

        args = np.load(args_file, allow_pickle=True)[0]

        if 'psf' in event_kwargs:
            args['psf'] = event_kwargs['psf'] in TRUE_OPTIONS

        for k in ['fcontam', 'min_sens', 'sys_err']:
            if k in event_kwargs:
                print('Set arg {0}={1}'.format(k, event_kwargs[k]))
                args[k] = event_kwargs[k]

        # Load MultiBeam
        mb = multifit.MultiBeam(beams_file, **args)

        if 'fit_trace_shift' in args:
            if args['fit_trace_shift']:
                tr = mb.fit_trace_shift()

        if 'spline_correction' in event_kwargs:
            spline_correction = event_kwargs['spline_correction'] in TRUE_OPTIONS
        else:
            spline_correction = True

        if 'fit_background' in event_kwargs:
            fit_background = event_kwargs['fit_background'] in TRUE_OPTIONS
        else:
            fit_background = True

        if 'fitter' in event_kwargs:
            fitter = event_kwargs['fitter']
        else:
            fitter = 'lstsq'

        if 'Rspline' in event_kwargs:
            Rspline = event_kwargs['Rspline']
        else:
            Rspline = 15

        if Rspline == 15:
            logg_list = [4.5]
        else:
            logg_list = utils.PHOENIX_LOGG

        if 'add_carbon_star' in event_kwargs:
            add_carbon_star = event_kwargs['add_carbon_star']
        else:
            add_carbon_star = 25

        if 'use_phoenix' in event_kwargs:
            p = event_kwargs.pop('use_phoenix')
            if p in TRUE_OPTIONS:
                tstar = utils.load_phoenix_stars(logg_list=logg_list,
                                             add_carbon_star=add_carbon_star)
            else:
                tstar = utils.load_templates(stars=True,
                                             add_carbon_star=add_carbon_star)
        else:
            tstar = utils.load_phoenix_stars(logg_list=logg_list,
                                             add_carbon_star=add_carbon_star)

        kws = {'spline_correction': spline_correction,
               'fit_background': fit_background,
               'fitter': fitter,
               'spline_args': {'Rspline': Rspline}}

        print('kwargs: {0}'.format(kws))

        # Fit the stellar templates
        _res = mb.xfit_star(tstar=tstar, oned_args={}, **kws)

        _res[0].savefig('{0}_{1:05d}.star.png'.format(root, id))

        # Save log info
        fp = open('{0}_{1:05d}.star.log'.format(root, id), 'w')
        fp.write(_res[1])
        fp.close()

        if output_path is None:
            # output_path = 'Pipeline/QuasarFit'.format(root)
            output_path = 'HST/Pipeline/{0}/Extractions'.format(root)

    else:
        # Normal galaxy redshift fit
        res = fitting.run_all_parallel(id, get_output_data=True,
                                       fit_only_beams=True, fit_beams=False,
                                       args_file=args_file,
                                       **event_kwargs)
        
        if root.startswith('fresco-g') | root.startswith('abell2744clu'):
            try:
                mb, tfit = res[0], res[3]
                # 1D for FRESCO
                fig1 = mb.oned_figure(bin=2, tfit=tfit,
                                      show_rest=False, minor=0.1,
                                      figsize=(9,3), median_filter_kwargs=None,
                                      ylim_percentile=0.1)
                
                hdu, fig2 = mb.drizzle_grisms_and_PAs(fcontam=0.0, 
                                                     flambda=False,
                                                     kernel='point', size=16,
                                                     tfit=tfit, diff=False, 
                                                      mask_segmentation=False,
                                    fig_args=dict(mask_segmentation=False, 
                                                  average_only=False,
                                                  scale_size=1, 
                                                  cmap='viridis_r', 
                                                  width_ratio=0.11),
                                                    )
            
                fig1.savefig(f'{root}_{id:05d}.1d.png')
                fig2.savefig(f'{root}_{id:05d}.2d.png')
            except:
                print('FRESCO outputs failed')
                
        if output_path is None:
            output_path = 'HST/Pipeline/{0}/Extractions'.format(root)

    # Output files
    files = glob.glob(f'{root}_{id:05d}*')
    
    for file in files:
        if ('beams.fits' not in file) | put_beams:
            aws_file = f'{output_path}/{file}'

            if event_kwargs['quasar_fit'] in TRUE_OPTIONS:
                # Don't copy stack
                if 'stack' in file:
                    continue

                # Add qso extension on outputs
                aws_file = aws_file.replace(f'_{id:05d}.', f'_{id:05d}.qso.')

            if update_db:
                print(f'Upload {file} -> {aws_file}')
                bkt.upload_file(file, aws_file,
                                ExtraArgs={'ACL': 'public-read'})
            else:
                print(f'Product {file}')
                
    # Put data in the redshift_fit database table
    try:
        if dbtable == 'stellar_fit':
            rowfile = f'{root}_{id:05d}.star.log'
        else:
            rowfile = f'{root}_{id:05d}.row.fits'

        if os.path.exists(rowfile) & update_db:
            grizli_db.add_redshift_fit_row(rowfile, table=dbtable,
                                           verbose=True)

    except:
        print('Update row failed')
        pass
    
    try:
        # Add 1D spectra
        files = glob.glob('{0}_{1:05d}*1D.fits'.format(root, id))
        if (len(files) > 0) & (dbtable == 'redshift_fit_v2') & update_db:
            grizli_db.send_1D_to_database(files=files)
    except:
        print(f'send_1D_to_database {files} failed')
        pass

    # Remove start log now that done
    if update_db:
        res = ubkt.delete_objects(Delete={'Objects': [{'Key': full_start}]})

    # Garbage collector
    gc.collect()


def clean(root='', verbose=True):
    import glob
    import os
    import matplotlib.pyplot as plt
    import gc

    plt.close('all')

    files = glob.glob(root+'*')
    files += glob.glob('*wcs.fits')
    files.sort()

    for file in files:
        print('Cleanup: {0}'.format(file))
        os.remove(file)

    gc.collect()


TESTER = 'HST/Pipeline/j001452+091221/Extractions/j001452+091221_00277.beams.fits'


def run_test(s3_object_path=TESTER):
    import os

    from importlib import reload
    import grizli.aws.lambda_handler
    reload(grizli.aws.lambda_handler)

    from grizli.aws.lambda_handler import run_grizli_fit, clean

    # event = {'s3_object_path': s3_object_path, 'verbose':'True'}

    obj = 'j224916m4432_02983'

    bucket = 'grizli'
    #root, id = obj.split('_')

    root = '_'.join(obj.split('_')[:-1])
    id = int(obj.split('_')[-1].split('.')[0])
    
    s3_object_path = 'HST/Pipeline/{0}/Extractions/{0}_{1:05d}.beams.fits'.format(root, int(id))
    event = {'s3_object_path': s3_object_path,
             'bucket': bucket,
             'verbose': 'True',
             'check_wcs': 'True',
             'zr': '0.1,0.3',
             }

    run_grizli_fit(event)

    beams_file = os.path.basename(event['s3_object_path'])
    #root = beams_file.split('_')[0]
    clean(root=root, verbose=True)


def handler(event, context):

    if 'handler_type' not in event:
        if 'scale_ab' in event:
            handler_type = 'drizzle_handler'
        else:
            handler_type = 'redshift_handler'
    else:
        handler_type = event.pop('handler_type')

    print('Handler type: ', handler_type)

    # Send the event to the appropriate function
    if handler_type == 'redshift_handler':
        redshift_handler(event, context)

    elif handler_type == 'show_version_handler':
        show_version(event, context)

    elif (handler_type == 'drizzle_handler'):
        from grizli.aws import aws_drizzler
        aws_drizzler.handler(event, context)


def redshift_handler(event, context):
    """
    Function for handling `run_grizli_fit` events
    """
    import matplotlib.pyplot as plt
    import traceback
    import time

    t0 = time.time()

    print(event)  # ['s3_object_path'], event['verbose'])
    # print(context)

    try:
        run_grizli_fit(event)
    except:
        print(traceback.format_exc(limit=2))
        pass

    # Clean up
    beams_file = os.path.basename(event['s3_object_path'])
    root = '_'.join(beams_file.split('_')[:-1])

    if 'clean' in event:
        run_clean = event['clean'] in TRUE_OPTIONS
    else:
        run_clean = True

    if run_clean:
        clean(root=root, verbose=True)

    t1 = time.time()

    return (event['s3_object_path'], t0, t1)


def show_version(event, context):
    import glob
    import grizli
    import eazy

    print('Event: ', event)

    print('grizli version: ', grizli.__version__)
    print('eazy version: ', eazy.__version__)

    import matplotlib
    print('matplotlibrc: ', matplotlib.matplotlib_fname())
    
    if 'files' in event:
        for file in event['files']:
            print('\n\n === File {0} === \n\n'.format(file))
            with open(file) as fp:
                lines = fp.readlines()
            
            print(''.join(lines))
    
    if 'glob' in event:
        _res = glob.glob(event['glob'])
        print('\n\n === glob {0} === \n\n'.format(event['glob']))
        for file in _res:
            print(file)
            
if __name__ == "__main__":
    handler('', '')



#!/bin/env python
import inspect
import os

import yaml

from collections import OrderedDict

"""
# shell
aws s3 ls s3://grizli-v1/Pipeline/j000200m5558/Prep/j000200m5558_visits.npy --request-payer requester

# Python
import boto3
s3 = boto3.resource('s3')
bkt = s3.Bucket('grizli-v1')
field = 'j000200m5558'
s3_file = '{0}_visits.npy'.format(field)
s3_path = 'Pipeline/{0}/Prep'.format(field)
bkt.download_file(s3_path+'/'+s3_file, s3_file,
              ExtraArgs={"RequestPayer": "requester"})
"""


class FilterDict(OrderedDict):
    meta = OrderedDict()

    @property
    def nfiles(self):
        """
        Count number of exposures
        """
        n = 0
        for k in self:
            n += len(self[k])
        return n

    @property
    def valid_filters(self):
        """
        Return a list of filters with N >= 1 files
        """
        valid = []
        for k in self:
            if len(self[k]) > 0:
                valid.append(k)
        return valid


def get_visit_files():
    import boto3
    from grizli.aws import db

    engine = db.get_db_engine()
    fields = db.from_sql("select field_root from charge_fields where log LIKE 'Finished%%'", engine=engine)

    s3 = boto3.resource('s3')
    bkt = s3.Bucket('grizli-v1')

    for i, field in enumerate(fields['field_root']):
        s3_file = '{0}_visits.npy'.format(field)
        if not os.path.exists(s3_file):
            s3_path = f'Pipeline/{field}/Prep'
            try:
                bkt.download_file(s3_path+'/'+s3_file, s3_file,
                              ExtraArgs={"RequestPayer": "requester"})
                print(i, s3_file)
            except:
                print(i, f'Download failed: {field}')
        else:
            print(f'Skip {field}')
    

def make_visit_fits():
    import glob
    import numpy as np
    from grizli import utils
    
    visit_files = glob.glob('[egu]*visits.npy')
    visit_files.sort()

    indiv_files = glob.glob('j*visits.npy')
    indiv_files.sort()

    visit_files += indiv_files

    for p in ['grizli-v1-19.12.04_visits.npy', 'grizli-v1-19.12.05_visits.npy', 'grizli-v1-20.10.12_visits.npy', 'grizli-cosmos-v2_visits.npy','grizli-v1-21.05.20_visits.npy']:
        if p in visit_files:
            visit_files.pop(visit_files.index(p))

    all_visits = []
    products = []

    extra_visits = ['candels-july2019_visits.npy', 'grizli-cosmos-v2_visits.npy']
    #extra_visits = ['candels-july2019_visits.npy', 'cosmos-dash-apr20_visits.npy']
    extra_visits = ['candels-july2019_visits.npy', 'cosmos-dash-dec06_visits.npy']
    
    for extra in extra_visits:

        extra_visits = np.load(extra, allow_pickle=True)[0]
        
        if 'cosmos-dash' in extra:
            extra_products = [v['product'] + '-'+v['files'][0][:6] for v in extra_visits]
        else:
            extra_products = [v['product'] for v in extra_visits]
        
        for i, p in enumerate(extra_products):
            if p not in products:
                parent = p.split('_')[0]
                #print(parent, p)
                v = extra_visits[i]
                v['parent'] = parent
                v['xproduct'] = v['product']
                v['parent_file'] = extra  # 'candels-july2019_visits.npy'
                all_visits.append(v)
                products.append(p)
            else:
                print('Skip: ', p, v['parent'])
                
    # COSMOS footprint
    cosmos_fp = None
    for i, v in enumerate(extra_visits):
        if v['product'].endswith('f814w'):
            print(v['product'])
            if cosmos_fp is None:
                cosmos_fp = v['footprint'].buffer(1.e-6)
            else:
                cosmos_fp = cosmos_fp.union(v['footprint'])

    for i, file in enumerate(visit_files):
        visits, groups, info = np.load(file, allow_pickle=True)
        print(file, len(visits))
        for v in visits:
            has_fp = ('footprints' in v)
            if not has_fp:
                print('No footprint: {0}'.format(v['product']))
            
            if file.startswith('j'):
                vprod = v['product'] + '-' + v['files'][0]
            else:
                vprod = v['product']
                
            if has_fp & (vprod not in products):
                v['parent'] = file.split("_visits")[0].split('-')[-1]
                v['first'] = v['files'][0]
                v['parent_file'] = file
                v['xproduct'] = vprod
                all_visits.append(v)
                products.append(vprod)

    for v in all_visits:
        v['filter'] = v['product'].split('-')[-1]
        v['first'] = v['files'][0]

    # File dictionary
    all_files = []
    file_products = []

    for v in all_visits:
        all_files.extend(v['files'])
        file_products.extend([v['xproduct']]*len(v['files']))

    # duplicates?? seem to be in GOODS-S.
    # Exclude them in all but the first product that contains them for now
    if True:
        _un = np.unique(all_files, return_counts=True, return_index=True, return_inverse=True)
        un_file, un_index, un_inv, un_count = _un
        dup = un_count > 1
        dup_files = un_file[dup]
        for file in dup_files:
            prods = list(np.array(file_products)[np.array(all_files) == file])
            for prod in prods[1:]:
                i = products.index(prod)
                v = all_visits[i]
                j = v['files'].index(file)
                print(file, v['parent'], prod, i, j)
                pj = all_visits[i]['files'].pop(j)
                pj = all_visits[i]['footprints'].pop(j)
                if 'awspath' in all_visits[i]:
                    pj = all_visits[i]['awspath'].pop(j)

            #print(file, prods[-1])

    # WFC3/IR copied to "Exposures" paths in CANDELS fields
    for v in all_visits:
        if v['parent_file'] in ['grizli-cosmos-v2_visits.npy', 'cosmos-dash-apr20_visits.npy', 'cosmos-dash-dec06_visits.npy']:
            continue

        if v['parent_file'].startswith('j'):
            v['awspath'] = ['grizli-v1/Pipeline/{0}/Prep'.format(v['parent']) for f in v['files']]
        else:
            #print(v['parent_file'], v['awspath'][0])
            
            if v['filter'].startswith('f0') | v['filter'].startswith('f1'):
                # print(v['product'])
                v['awspath'] = ['grizli-v1/Exposures/{0}/{1}'.format(f[:4], f.split('_')[0]) for f in v['files']]

    # Empty visits, seems to be from duplicates above and mostly in CANDELS
    nexp = np.array([len(visit['files']) for visit in all_visits])
    for i in np.where(nexp == 0)[0][::-1]:
        v_i = all_visits.pop(i)
        print(i, v_i['product'])
        products.pop(i)

    tab = utils.GTable()

    for k in ['parent', 'product', 'filter', 'first']:
        tab[k] = [visit[k] for visit in all_visits]

    coo = np.array([np.array(visit['footprint'].centroid.xy).flatten() for visit in all_visits])
    tab['ra'] = coo[:, 0]
    tab['dec'] = coo[:, 1]
    tab['nexp'] = [len(visit['files']) for visit in all_visits]
    tab['bounds'] = [np.array(v['footprint'].bounds) for v in all_visits]

    root = 'candels-july2019'
    root = 'candels-sep2019'
    root = 'grizli-v1-19.12.04'
    root = 'grizli-v1-19.12.05'
    root = 'grizli-v1-20.10.12'
    root = 'grizli-v1-21.05.20'
    root = 'grizli-v1-21.12.18'

    tab.write(root+'_visits.fits', overwrite=True)
    np.save(root+'_visits.npy', [all_visits])

    # os.system('echo "# In https://s3.amazonaws.com/grizli-v1/Mosaics/" > candels-july2019.files.txt; ls candels-july2019* |grep -v files.txt >>  candels-july2019.files.txt')
    #os.system('aws s3 sync --exclude "*" --include "candels-july2019*" --include "grizli-v1-19.12.04*" ./ s3://grizli-v1/Mosaics/ --acl public-read')
    os.system('echo "# In https://s3.amazonaws.com/grizli-v1/Mosaics/" > {0}.files.txt; ls {0}* |grep -v files.txt >>  {0}.files.txt'.format(root))

    os.system('aws s3 sync --exclude "*" --include "{0}*" ./ s3://grizli-v1/Mosaics/ --acl public-read'.format(root))

    if False:
        from shapely.geometry import Point
        from grizli.aws import db
        engine = db.get_db_engine() 
        
        fields = db.from_sql("select field_root, a_wfc3 from charge_fields where log LIKE 'Finished%%'", engine=engine)
        
        candels = utils.column_values_in_list(tab['parent'], ['j141956p5255', 'j123656p6215', 'j033236m2748', 'j021732m0512', 'j100012p0210'])
        cosmos = np.array([v['footprint'].intersection(cosmos_fp).area > 0 for v in all_visits])

        extra = candels
        extra = ~(candels | cosmos)

        # Area
        filter_polys = {}

        filt = 'f160w'
        for filt in np.unique(tab['filter']):
            print(filt)
            if filt in filter_polys:
                print(filt)
                continue

            poly = None
            count = 0

            # Dec strips
            di = np.arange(-90, 91, 5)
            strips = []
            for i in range(len(di)-1):
                strip = (tab['dec'] > di[i]) & (tab['dec'] <= di[i+1]) & (tab['filter'] == filt)
                strip &= extra

                if strip.sum() == 0:
                    continue

                indices = np.arange(len(tab))[strip]

                poly = None
                for j in indices:
                    v = all_visits[j]
                    if v['filter'] != filt:
                        continue

                    # for fp in v['footprints']:
                    for fp in [v['footprint']]:
                        count += 1
                        #print(i, v['product'], count)
                        if poly is None:
                            poly = fp.buffer(1.e-6)
                        else:
                            poly = poly.union(fp.buffer(1.e-6))

                poly.dec = di[i]+2.5
                strips.append(poly)

            if len(strips) == 0:
                filter_polys[filt] = Point(0, 0).buffer(1.e-6)
                continue

            full = strips[0].buffer(1.e-6)
            for strip in strips[1:]:
                full = full.union(strip.buffer(1.e-6))

            filter_polys[filt] = full

        optical = filter_polys['f606w'].union(filter_polys['f814w'])
        optical = optical.union(filter_polys['f850lp'])
        optical = optical.union(filter_polys['f775w'])

        yband = filter_polys['f098m'].union(filter_polys['f105w'])

        visy = optical.union(yband)

        jband = filter_polys['f125w']
        jband = jband.union(filter_polys['f110w'])

        hband = filter_polys['f140w'].union(filter_polys['f160w'])

        filter_polys[r'$\mathrm{opt} = i_{775} | i_{814} | z_{850}$'] = optical
        filter_polys[r'$\mathrm{opty} = \mathrm{opt} | Y$'] = visy
        filter_polys[r'$Y = y_{098 } | y_{105}$'] = yband
        filter_polys[r'$J = j_{110} | j_{125}$'] = jband
        filter_polys[r'$H = h_{140} | h_{160}$'] = hband

        ydrop = visy.intersection(jband)
        ydrop = ydrop.intersection(hband)
        filter_polys[r'$Y-\mathrm{drop} = (\mathrm{opt} | Y) + J + H$'] = ydrop

        yj = yband.union(jband)
        jdrop = yj.intersection(hband)
        filter_polys[r'$J-\mathrm{drop} = (Y | J) + H$'] = jdrop

        for filt in filter_polys:
            full = filter_polys[filt]
            try:
                areas = [f.area*np.cos(np.array(f.centroid.xy).flatten()[1]/180*np.pi) for f in full]
            except:
                try:
                    areas = [f.area*np.cos(np.array(f.centroid.xy).flatten()[1]/180*np.pi) for f in [full]]
                except:
                    areas = [0]

            full.total_area = np.sum(areas)
            print(filt, filter_polys[filt].total_area)

        ta = utils.GTable()
        ta['filter'] = [f.upper() for f in filter_polys]
        ta['area'] = [filter_polys[f].total_area*3600 for f in filter_polys]
        ta['area'].format = '.0f'

        # Compare areas
        h = fields['a_wfc3_ir_f160w'] > 0
        for root, aa in zip(fields['field_root'][h], fields['a_wfc3_ir_f160w'][h]):
            sel = (tab['filter'] == 'f160w') & (tab['parent'] == root)
            if sel.sum() > 0:
                indices = np.where(sel)[0]
                a = all_visits[indices[0]]['footprint'].buffer(1.e-6)
                for i in indices:
                    a = a.union(all_visits[i]['footprint'])

                a_i = a.area*3600*np.cos(tab['dec'][indices[0]]/180*np.pi)
                print(root, aa, a_i, a_i/aa)

def update_s3_paths():
    """
    Update paths to COSMOS data, eventually everything else
    """
    import numpy as np
    import os
    from tqdm import tqdm
    
    os.chdir(os.path.join(os.getenv('HOME'),
                     'Research/HST/CHArGE/Cutouts/VisitFiles'))
    
    groups = np.load('grizli-v1-21.12.18_filter_groups.npy', allow_pickle=True)
    
    for f in groups[0]:
        print(f)
        for j, aws in tqdm(enumerate(groups[0][f]['awspath'])):
            groups[0][f]['awspath'][j] = aws.replace('cosmos-dash',
                                                     'cosmos-dash-2021')
    
    np.save('grizli-v2-22.02.02_filter_groups.npy', groups)
                                     
def group_by_filter():
    """
    aws s3 sync --exclude "*" --include "cosmos_visits*" s3://grizli-preprocess/CosmosMosaic/ ./

    """
    from grizli import prep, utils
    import numpy as np

    master = 'cosmos'
    master = 'grizli-cosmos-v2'
    master = 'grizli-jan2019'
    master = 'grizli-v1-19.12.04'
    master = 'grizli-v1-19.12.05'
    master = 'grizli-v1-20.10.12'
    master = 'grizli-v1-21.05.20'
    master = 'grizli-v1-21.12.18'
    
    tab = utils.read_catalog('{0}_visits.fits'.format(master))
    all_visits = np.load('{0}_visits.npy'.format(master), allow_pickle=True)[0]

    # By filter

    # Exclude DASH
    dash = utils.column_string_operation(tab['product'], 'icxe', 'startswith')
    dash |= utils.column_string_operation(tab['product'], '_icxe',
                                         'count', 'or')

    # Don't exclude DASH
    dash = utils.column_string_operation(tab['product'], 'xxxx', 'startswith')

    groups = {}
    fpstr = {}

    for filt in np.unique(tab['filter']):
        mat = (tab['filter'] == filt) & (~dash)
        groups[filt] = {'filter': filt, 'files': [], 'awspath': [], 'footprints': []}
        fpstr[filt] = 'fk5\n'

        for ix in np.where(mat)[0]:
            fp = all_visits[ix]['footprint']
            if hasattr(fp, '__len__'):
                fps = fp
            else:
                fps = [fp]
            for fp in fps:
                xy = fp.boundary.xy
                pstr = 'polygon('+','.join(['{0:.6f}'.format(i) for i in np.array([xy[0].tolist(), xy[1].tolist()]).T.flatten()])+') # text={{{0}}}\n'.format(all_visits[ix]['product'])

                fpstr[filt] += pstr

            for k in ['files', 'awspath', 'footprints']:
                groups[filt][k].extend(all_visits[ix][k])

        fp = open('{0}-pointings-{1}.reg'.format(master, filt), 'w')
        fp.write(fpstr[filt])
        fp.close()

        print('{0:6} {1:>3d} {2:>4d} ({3:>4d})'.format(filt, mat.sum(), len(groups[filt]['files']), len(np.unique(groups[filt]['files']))))

    np.save('{0}_filter_groups.npy'.format(master), [groups])

    os.system('aws s3 sync --exclude "*" --include "{0}*" ./ s3://grizli-v1/Mosaics/ --acl public-read'.format(master))

# RGB_PARAMS = {'xsize':4, 'rgb_min':-0.01, 'verbose':True, 'output_dpi': None, 'add_labels':False, 'output_format':'png', 'show_ir':False, 'scl':2, 'suffix':'.rgb', 'mask_empty':False}


RGB_PARAMS = {'xsize': 4,
              'output_dpi': None,
              'rgb_min': -0.01,
              'add_labels': False,
              'output_format': 'png',
              'show_ir': False,
              'scl': 2,
              'suffix': '.rgb',
              'mask_empty': False,
              'tick_interval': 1,
              'pl': 1,  # 1 for f_lambda, 2 for f_nu
              }

# xsize=4, output_dpi=None, HOME_PATH=None, show_ir=False, pl=1, pf=1, scl=1, rgb_scl=[1, 1, 1], ds9=None, force_ir=False, filters=all_filters, add_labels=False, output_format='png', rgb_min=-0.01, xyslice=None, pure_sort=False, verbose=True, force_rgb=None, suffix='.rgb', scale_ab=scale_ab)


def segmentation_figure(label, cat, segfile):
    """
    Make a figure showing a cutout of the segmentation file
    """
    import matplotlib.pyplot as plt
    import numpy as np

    import astropy.io.fits as pyfits
    import astropy.wcs as pywcs
    from grizli import utils

    plt.ioff()

    seg = pyfits.open(segfile)
    seg_data = seg[0].data
    seg_wcs = pywcs.WCS(seg[0].header)

    # Randomize seg to get dispersion between neighboring objects
    np.random.seed(hash(label.split('_')[0]) % (10 ** 8))
    rnd_ids = np.append([0], np.argsort(np.random.rand(len(cat)))+1)

    # Make cutout
    th = pyfits.open('{0}.thumb.fits'.format(label), mode='update')
    th_wcs = pywcs.WCS(th[0].header)
    blot_seg = utils.blot_nearest_exact(seg_data, seg_wcs, th_wcs,
                               stepsize=-1, scale_by_pixel_area=False)

    rnd_seg = rnd_ids[np.asarray(blot_seg,dtype=int)]*1.
    th_ids = np.unique(blot_seg)

    sh = th[0].data.shape
    yp, xp = np.indices(sh)

    thumb_height = 2.
    fig = plt.figure(figsize=[thumb_height*sh[1]/sh[0], thumb_height])
    ax = fig.add_subplot(111)
    rnd_seg[rnd_seg == 0] = np.nan

    ax.imshow(rnd_seg, aspect='equal', cmap='terrain_r',
              vmin=-0.05*len(cat), vmax=1.05*len(cat))
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    ix = utils.column_values_in_list(cat['number'], th_ids)
    xc, yc = th_wcs.all_world2pix(cat['ra'][ix], cat['dec'][ix], 0)
    xc = np.clip(xc, 0.09*sh[1], 0.91*sh[1])
    yc = np.clip(yc, 0.08*sh[0], 0.92*sh[0])

    for th_id, x_i, y_i in zip(cat['number'][ix], xc, yc):
        if th_id == 0:
            continue

        ax.text(x_i, y_i, '{0:.0f}'.format(th_id), ha='center', va='center', fontsize=8,  color='w')
        ax.text(x_i, y_i, '{0:.0f}'.format(th_id), ha='center', va='center', fontsize=8,  color='k', alpha=0.95)

    ax.set_xlim(0, sh[1]-1)
    ax.set_ylim(0, sh[0]-1)
    ax.set_axis_off()

    fig.tight_layout(pad=0.01)
    fig.savefig('{0}.seg.png'.format(label))
    plt.close(fig)

    # Append to thumbs file
    seg_hdu = pyfits.ImageHDU(data=np.asarray(blot_seg,dtype=int), name='SEG')
    if 'SEG' in th:
        th.pop('SEG')

    th.append(seg_hdu)
    th.writeto('{0}.thumb.fits'.format(label), overwrite=True,
                 output_verify='fix')
    th.close()


def drizzle_images(label='macs0647-jd1', ra=101.9822125, dec=70.24326667, pixscale=0.1, size=10, wcs=None, pixfrac=0.33, kernel='square', theta=0, half_optical_pixscale=True, filters=['f160w', 'f140w', 'f125w', 'f105w', 'f110w', 'f098m', 'f850lp', 'f814w', 'f775w', 'f606w', 'f475w', 'f555w', 'f600lp', 'f390w', 'f350lp'], weight_type="jwst_var", skip=None, remove=True, rgb_params=RGB_PARAMS, master='grizli-v1-19.12.04', aws_bucket='s3://grizli/CutoutProducts/', scale_ab=21, thumb_height=2.0, sync_fits=True, subtract_median=True, include_saturated=True, include_ir_psf=False, oversample_psf=False, show_filters=['visb', 'visr', 'y', 'j', 'h'], combine_similar_filters=True, single_output=True, aws_prep_dir=None, make_segmentation_figure=False, get_dict=False, dryrun=False, thumbnail_ext='png', write_ctx=False, **kwargs):
    """
    label='cp561356'; ra=150.208875; dec=1.850241667; size=40; filters=['f160w','f814w', 'f140w','f125w','f105w','f606w','f475w']

    master: These are sets of large lists of available exposures

        'cosmos': deprecated
        'grizli-cosmos-v2': All imaging covering the COSMOS field
        'candels-july2019': CANDELS fields other than COSMOS
        'grizli-v1': First processing of the Grizli CHArGE dataset
        'grizli-v1-19.12.04': Updated CHArGE fields
            ** this is now a copy from 21.05.20 so that the old lambda
               function can catch it **

        'grizli-v1-21.05.20': ACS fields + new cosmos

    """
    import glob
    import copy
    import os

    import numpy as np

    import astropy.io.fits as pyfits
    from astropy.coordinates import SkyCoord
    import astropy.units as u
    from drizzlepac.adrizzle import do_driz

    import boto3

    from grizli import prep, utils
    from grizli.pipeline import auto_script

    # Function arguments
    if get_dict:
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame).locals

        pop_args = ['get_dict', 'frame', 'kwargs']
        pop_classes = (np.__class__, do_driz.__class__, SkyCoord.__class__)

        for k in kwargs:
            args[k] = kwargs[k]

        for k in args:
            if isinstance(args[k], pop_classes):
                pop_args.append(k)

        for k in pop_args:
            if k in args:
                args.pop(k)

        return args

    # Boto objects
    s3 = boto3.resource('s3')
    s3_client = boto3.client('s3')

    if isinstance(ra, str):
        coo = SkyCoord('{0} {1}'.format(ra, dec), unit=(u.hour, u.deg))
        ra, dec = coo.ra.value, coo.dec.value

    if label is None:
        try:
            import mastquery.utils
            label = mastquery.utils.radec_to_targname(ra=ra, dec=dec, round_arcsec=(1/15, 1), targstr='j{rah}{ram}{ras}{sign}{ded}{dem}{des}')
        except:
            label = 'grizli-cutout'

    #master = 'cosmos'
    #master = 'grizli-jan2019'

    if master == 'grizli-jan2019':
        parent = 's3://grizli/MosaicTools/'
        bkt = s3.Bucket('grizli')
    elif master == 'cosmos':
        parent = 's3://grizli-preprocess/CosmosMosaic/'
        bkt = s3.Bucket('grizli-preprocess')
    elif master == 'grizli-cosmos-v2':
        parent = 's3://grizli-cosmos-v2/Mosaics/'
        bkt = s3.Bucket('grizli-cosmos-v2')
    elif master == 'candels-july2019':
        parent = 's3://grizli-v1/Mosaics/'
        bkt = s3.Bucket('grizli-v1')
    elif master == 'grizli-v1-19.12.04':
        parent = 's3://grizli-v1/Mosaics/'
        bkt = s3.Bucket('grizli-v1')
    elif master == 'grizli-v1-19.12.05':
        parent = 's3://grizli-v1/Mosaics/'
        bkt = s3.Bucket('grizli-v1')
    elif master == 'grizli-v1-latest':
        parent = 's3://grizli-v1/Mosaics/'
        bkt = s3.Bucket('grizli-v1')        
    else:
        # Run on local files, e.g., "Prep" directory
        parent = None
        bkt = None
        #remove = False

    # Download summary files from S3
    for ext in ['_visits.fits', '_visits.npy', '_filter_groups.npy'][-1:]:
        newfile = '{0}{1}'.format(master, ext)
        if (not os.path.exists(newfile)) & (parent is not None):

            s3_path = parent.split('/')[-2]
            s3_file = '{0}{1}'.format(master, ext)
            print('{0}{1}'.format(parent, s3_file))
            bkt.download_file(s3_path+'/'+s3_file, s3_file,
                              ExtraArgs={"RequestPayer": "requester"})

            #os.system('aws s3 cp {0}{1}{2} ./'.format(parent, master, ext))

    #tab = utils.read_catalog('{0}_visits.fits'.format(master))
    #all_visits = np.load('{0}_visits.npy'.format(master))[0]
    if parent is not None:
        groups = np.load('{0}_filter_groups.npy'.format(master), allow_pickle=True)[0]
    else:

        if aws_prep_dir is not None:
            spl = aws_prep_dir.replace('s3://', '').split('/')
            prep_bucket = spl[0]
            prep_root = spl[2]

            prep_bkt = s3.Bucket(prep_bucket)

            s3_prep_path = 'Pipeline/{0}/Prep/'.format(prep_root)
            s3_full_path = '{0}/{1}'.format(prep_bucket, s3_prep_path)
            s3_file = '{0}_visits.npy'.format(prep_root)

            # Make output path Prep/../Thumbnails/
            if aws_bucket is not None:
                aws_bucket = ('s3://' +
                              s3_full_path.replace('/Prep/', '/Thumbnails/'))

            print('{0}{1}'.format(s3_prep_path, s3_file))
            if not os.path.exists(s3_file):
                prep_bkt.download_file(os.path.join(s3_prep_path, s3_file),
                            s3_file, ExtraArgs={"RequestPayer": "requester"})

            groups_files = glob.glob('{0}_filter_groups.npy'.format(prep_root))
            visit_query = prep_root+'_'
        else:
            groups_files = glob.glob('*filter_groups.*')
            visit_query = '*'

        # Reformat local visits.npy into a groups file
        if (len(groups_files) == 0):

            visit_file = glob.glob(visit_query+'visits.*')[0]
            visit_root = visit_file.split('_visits')[0]
            visits, groups, info = auto_script.load_visit_info(visit_root, 
                                                               verbose=False)

            #visits, groups, info = np.load(visit_file, allow_pickle=True)
            visit_filters = np.array([v['product'].split('-')[-1]
                                      for v in visits])
                                      
            groups = {}
            sgroups = {}
            
            for filt in np.unique(visit_filters):
                _v = {}
                _v['filter'] = str(filt)
                _v['files'] = []
                _v['footprints'] = []
                _v['awspath'] = []

                ix = np.where(visit_filters == filt)[0]
                for i in ix:
                    _v['files'].extend(visits[i]['files'])
                    _v['footprints'].extend(visits[i]['footprints'])

                Nf = len(_v['files'])
                print('{0:>6}: {1:>3} exposures'.format(filt, Nf))

                if aws_prep_dir is not None:
                    _v['awspath'] = [s3_full_path for file in range(Nf)]
            
                groups[filt] = _v 
                sgroups[str(filt)] = auto_script.visit_dict_to_strings(_v)
                
            #np.save('{0}_filter_groups.npy'.format(visit_root), [groups])
            with open('{0}_filter_groups.yaml'.format(visit_root), 'w') as fp:
                yaml.dump(sgroups, stream=fp, Dumper=yaml.Dumper)
        else:
            print('Use groups file: {0}'.format(groups_files[0]))

            #groups = np.load(groups_files[0], allow_pickle=True)[0]
            with open(groups_files[0]) as fp:
                groups = yaml.load(fp, Loader=yaml.Loader)
                for filt in groups:
                    _v = groups[filt]
                    groups[filt] = auto_script.visit_dict_from_strings(_v)
                    
    #filters = ['f160w','f814w', 'f110w', 'f098m', 'f140w','f125w','f105w','f606w', 'f475w']

    filt_dict = FilterDict()
    filt_dict.meta['label'] = label
    filt_dict.meta['ra'] = ra
    filt_dict.meta['dec'] = dec
    filt_dict.meta['size'] = size
    filt_dict.meta['master'] = master
    filt_dict.meta['parent'] = parent

    if filters is None:
        filters = list(groups.keys())

    has_filts = []
    lower_filters = [f.lower() for f in filters]
    for filt in lower_filters:
        if filt not in groups:
            continue

        visits = [copy.deepcopy(groups[filt])]
        #visits[0]['reference'] = 'CarlosGG/ak03_j1000p0228/Prep/ak03_j1000p0228-f160w_drz_sci.fits'

        visits[0]['product'] = label+'-'+filt

        if wcs is None:
            hdu = utils.make_wcsheader(ra=ra, dec=dec, size=size, pixscale=pixscale, get_hdu=True, theta=theta)

            h = hdu.header
        else:
            h = utils.to_header(wcs)

        if (filt[:2] in ['f0', 'f1', 'g1']) | (not half_optical_pixscale):
            #data = hdu.data
            pass
        else:
            for k in ['NAXIS1', 'NAXIS2', 'CRPIX1', 'CRPIX2']:
                h[k] *= 2

            h['CRPIX1'] -= 0.5
            h['CRPIX2'] -= 0.5

            for k in ['CD1_1', 'CD1_2', 'CD2_1', 'CD2_2']:
                if k in h:
                    h[k] /= 2

            #data = np.zeros((h['NAXIS2'], h['NAXIS1']), dtype=np.int16)

        #pyfits.PrimaryHDU(header=h, data=data).writeto('ref.fits', overwrite=True, output_verify='fix')
        #visits[0]['reference'] = 'ref.fits'

        print('\n\n###\nMake filter: {0}'.format(filt))

        if include_ir_psf:
            clean_i = False
        else:
            clean_i = remove

        status = utils.drizzle_from_visit(
            visits[0],
            h,
            pixfrac=pixfrac,
            kernel=kernel,
            clean=clean_i,
            include_saturated=include_saturated,
            weight_type=weight_type,
            skip=skip,
            dryrun=dryrun
        )

        if dryrun:
            filt_dict[filt] = status
            continue

        elif status is not None:
            sci, wht, var, ctx, outh, filt_dict[filt], wcs_tab = status

            if subtract_median:
                #med = np.median(sci[sci != 0])
                try:
                    un_data = np.unique(sci[(sci != 0) & np.isfinite(sci)])
                    med = utils.mode_statistic(un_data)
                except:
                    med = 0.

                if not np.isfinite(med):
                    med = 0.

                print('\n\nMedian {0} = {1:.3f}\n\n'.format(filt, med))
                outh['IMGMED'] = (med, 'Median subtracted from the image')
            else:
                med = 0.
                outh['IMGMED'] = (0., 'Median subtracted from the image')

            # Write files
            pyfits.writeto(
                '{0}-{1}_drz_sci.fits'.format(label, filt),
                 data=sci,
                 header=outh,
                 overwrite=True,
                 output_verify='fix'
            )

            pyfits.writeto(
                '{0}-{1}_drz_wht.fits'.format(label, filt),
                data=wht,
                header=outh,
                overwrite=True,
                output_verify='fix'
            )

            if var is not None:
                pyfits.writeto(
                    '{0}-{1}_drz_var.fits'.format(label, filt),
                    data=var,
                    header=outh,
                    overwrite=True,
                    output_verify='fix'
                )

            if write_ctx:
                pyfits.writeto(
                    '{0}-{1}_drz_ctx.fits'.format(label, filt),
                    data=ctx,
                    header=outh,
                    overwrite=True,
                    output_verify='fix'
                )

            has_filts.append(filt)

            if include_ir_psf:
                from grizli.galfit.psf import DrizzlePSF

                hdu = pyfits.open('{0}-{1}_drz_sci.fits'.format(label, filt),
                                  mode='update')

                flt_files = []  # visits[0]['files']
                for i in range(1, 10000):
                    key = 'FLT{0:05d}'.format(i)
                    if key not in hdu[0].header:
                        break

                    flt_files.append(hdu[0].header[key])

                try:

                    dp = DrizzlePSF(flt_files=flt_files, driz_hdu=hdu[0])
                    
                    if oversample_psf:
                        oN = oversample_psf*2+1
                        cosd = np.cos(dp.driz_wcs.wcs.crval[1]/180*np.pi)
                        dde = 1./(oversample_psf*2)*pixscale/3600
                        dra = dde*cosd
                        sh = sci.shape
                        psfd = np.zeros((oN*sh[0], oN*sh[1]), 
                                        dtype=np.float32)
                        for i in range(oN):
                            for j in range(oN):
                                ra_i = (dp.driz_wcs.wcs.crval[0] +
                                        dra*(i-oversample_psf))
                                de_i = (dp.driz_wcs.wcs.crval[1] - 
                                        dde*(j-oversample_psf))
                                psf_i = dp.get_psf(ra=ra_i, dec=de_i,
                                         filter=filt.upper(),
                                         pixfrac=dp.driz_header['PIXFRAC'],
                                         kernel=dp.driz_header['KERNEL'],
                                         wcs_slice=dp.driz_wcs, 
                                get_extended=filt.lower()[:2] in ['f1','f0'],
                                         verbose=False, get_weight=False)
                                psfd[j::oN,i::oN] += psf_i[1].data
                        
                        psf = pyfits.ImageHDU(data=psfd)
                    else:
                        psf = dp.get_psf(ra=dp.driz_wcs.wcs.crval[0],
                                 dec=dp.driz_wcs.wcs.crval[1],
                                 filter=filt.upper(),
                                 pixfrac=dp.driz_header['PIXFRAC'],
                                 kernel=dp.driz_header['KERNEL'],
                                 wcs_slice=dp.driz_wcs, 
                                 get_extended=filt.lower()[:2] in ['f1','f0'],
                                 verbose=False, get_weight=False)[1]
                    
                    psf.header['OVERSAMP'] = oversample_psf
                    
                    psf.header['EXTNAME'] = 'PSF'
                    #psf[1].header['EXTVER'] = filt
                    hdu.append(psf)
                    hdu.flush()

                except:
                    pass

        if remove:
            os.system('rm *_fl*fits')

    # Dry run, just return dictionary of the found exposure files
    if dryrun:
        return filt_dict

    # Nothing found
    if len(has_filts) == 0:
        return []

    if combine_similar_filters:
        combine_filters(label=label)

    if rgb_params:
        #auto_script.field_rgb(root=label, HOME_PATH=None, filters=has_filts, **rgb_params)
        show_all_thumbnails(label=label, thumb_height=thumb_height, scale_ab=scale_ab, close=True, rgb_params=rgb_params, filters=show_filters, ext=thumbnail_ext)

    if (single_output != 0):
        # Concatenate into a single FITS file
        files = glob.glob('{0}-f*_dr[cz]_sci.fits'.format(label))
        files.sort()

        if combine_similar_filters:
            comb_files = glob.glob('{0}-[a-eg-z]*_dr[cz]_sci.fits'.format(label))
            comb_files.sort()
            files += comb_files

        hdul = None
        for file in files:
            hdu_i = pyfits.open(file)
            hdu_i[0].header['EXTNAME'] = 'SCI'
            if 'NCOMBINE' in hdu_i[0].header:
                if hdu_i[0].header['NCOMBINE'] <= single_output:
                    continue

                filt_i = file.split('-')[-1].split('_dr')[0]
            else:
                filt_i = utils.parse_filter_from_header(hdu_i[0].header)

            for h in hdu_i:
                h.header['EXTVER'] = filt_i
                if hdul is None:
                    hdul = pyfits.HDUList([h])
                else:
                    hdul.append(h)

            print('Add to {0}.thumb.fits: {1}'.format(label, file))

            # Weight
            hdu_i = pyfits.open(file.replace('_sci', '_wht'))
            hdu_i[0].header['EXTNAME'] = 'WHT'
            for h in hdu_i:
                h.header['EXTVER'] = filt_i
                if hdul is None:
                    hdul = pyfits.HDUList([h])
                else:
                    hdul.append(h)

        hdul.writeto('{0}.thumb.fits'.format(label), overwrite=True,
                     output_verify='fix')

        for file in files:
            for f in [file, file.replace('_sci', '_wht')]:
                if os.path.exists(f):
                    print('Remove {0}'.format(f))
                    os.remove(f)

    # Segmentation figure
    thumb_file = '{0}.thumb.fits'.format(label)
    if (make_segmentation_figure) & (os.path.exists(thumb_file)) & (aws_prep_dir is not None):

        print('Make segmentation figure')

        # Fetch segmentation image and catalog
        s3_prep_path = 'Pipeline/{0}/Prep/'.format(prep_root)
        s3_full_path = '{0}/{1}'.format(prep_bucket, s3_prep_path)
        s3_file = '{0}_visits.npy'.format(prep_root)

        has_seg_files = True
        seg_files = ['{0}-ir_seg.fits.gz'.format(prep_root),
                     '{0}_phot.fits'.format(prep_root)]

        for s3_file in seg_files:
            if not os.path.exists(s3_file):
                remote_file = os.path.join(s3_prep_path, s3_file)
                try:
                    print('Fetch {0}'.format(remote_file))
                    prep_bkt.download_file(remote_file, s3_file,
                                   ExtraArgs={"RequestPayer": "requester"})
                except:
                    has_seg_files = False
                    print('Make segmentation figure failed: {0}'.format(remote_file))
                    break

        if has_seg_files:
            s3_cat = utils.read_catalog(seg_files[1])
            segmentation_figure(label, s3_cat, seg_files[0])

    if aws_bucket:
        #aws_bucket = 's3://grizli-cosmos/CutoutProducts/'
        #aws_bucket = 's3://grizli/CutoutProducts/'

        s3 = boto3.resource('s3')
        s3_client = boto3.client('s3')
        bkt = s3.Bucket(aws_bucket.split("/")[2])
        aws_path = '/'.join(aws_bucket.split("/")[3:])

        if sync_fits:
            files = glob.glob('{0}*'.format(label))
        else:
            files = glob.glob('{0}*png'.format(label))

        for file in files:
            print('{0} -> {1}'.format(file, aws_bucket))
            bkt.upload_file(file, '{0}/{1}'.format(aws_path, file).replace('//', '/'), ExtraArgs={'ACL': 'public-read'})

        #os.system('aws s3 sync --exclude "*" --include "{0}*" ./ {1} --acl public-read'.format(label, aws_bucket))

        #os.system("""echo "<pre>" > index.html; aws s3 ls AWSBUCKETX --human-readable | sort -k 1 -k 2 | grep -v index | awk '{printf("%s %s",$1, $2); printf(" %6s %s ", $3, $4); print "<a href="$5">"$5"</a>"}'>> index.html; aws s3 cp index.html AWSBUCKETX --acl public-read""".replace('AWSBUCKETX', aws_bucket))

    return has_filts


def get_cutout_from_aws(label='macs0647-jd1', ra=101.9822125, dec=70.24326667, master='grizli-jan2019', scale_ab=21, thumb_height=2.0, remove=1, aws_bucket="s3://grizli/DropoutThumbnails/", lambda_func='grizliImagingCutout', force=False, **kwargs):
    """
    Get cutout using AWS lambda
    """
    import boto3
    import json

    #func = 'grizliImagingCutout'

    #label = '{0}_{1:05d}'.format(self.cat['root'][ix], self.cat['id'][ix])
    #url = 'https://s3.amazonaws.com/grizli/DropoutThumbnails/{0}.thumb.png'

    session = boto3.Session()
    client = session.client('lambda', region_name='us-east-1')

    event = {
          'label': label,
          "ra": ra,
          "dec": dec,
          "scale_ab": scale_ab,
          "thumb_height": thumb_height,
          "aws_bucket": aws_bucket,
          "remove": remove,
          "master": master,
        }

    for k in kwargs:
        event[k] = kwargs[k]

    bucket_split = aws_bucket.strip("s3://").split('/')
    bucket_name = bucket_split[0]
    bucket_path = '/'.join(bucket_split[1:])

    s3 = boto3.resource('s3')
    s3_client = boto3.client('s3')
    bkt = s3.Bucket(bucket_name)

    files = [obj.key for obj in bkt.objects.filter(Prefix='{0}/{1}.thumb.png'.format(bucket_path, label))]

    if (len(files) == 0) | force:
        print('Call lambda: {0}'.format(label))

        print(event)

        response = client.invoke(
            FunctionName=lambda_func,
            InvocationType='Event',
            LogType='Tail',
            Payload=json.dumps(event))
    else:
        response = None
        print('Thumb exists')

    return response


def handler(event, context):
    import os
    import grizli
    print(grizli.__version__)

    os.chdir('/tmp/')
    os.system('rm *')
    os.system('rm -rf matplotlib*')

    print(event)  # ['s3_object_path'], event['verbose'])
    drizzle_images(**event)
    os.system('rm *')


def combine_filters(label='j022708p4901_00273', verbose=True):
    """
    Group nearby filters
    """
    import glob
    import numpy as np
    import astropy.io.fits as pyfits
    from grizli import utils

    filter_queries = {}
    filter_queries['uv'] = '{0}-f[2-3]*sci.fits'.format(label)
    filter_queries['visb'] = '{0}-f[4-5]*sci.fits'.format(label)
    filter_queries['visr'] = '{0}-f[6-8]*sci.fits'.format(label)
    filter_queries['y'] = '{0}-f[01][90][85]*sci.fits'.format(label)
    filter_queries['j'] = '{0}-f1[12][05]*sci.fits'.format(label)
    filter_queries['h'] = '{0}-f1[64]0*sci.fits'.format(label)

    grouped_filters = {}

    for qfilt in filter_queries:

        drz_files = glob.glob(filter_queries[qfilt])
        drz_files.sort()
        grouped_filters[qfilt] = [f.split('_dr')[0].split('-')[-1] for f in drz_files]

        if len(drz_files) > 0:
            drz_files.sort()

            if verbose:
                print('# Combine filters, {0}={1}'.format(qfilt,
                                      '+'.join(drz_files)))

            for i, file in enumerate(drz_files[::-1]):
                drz = pyfits.open(file)
                wht = pyfits.open(file.replace('_sci', '_wht'))
                sci = drz[0].data*1.

                # Subtract background?
                if 'IMGMED' in drz[0].header:
                    sci -= drz[0].header['IMGMED']
                    drz[0].header['IMGMED'] = 0.

                if i == 0:
                    photflam = drz[0].header['PHOTFLAM']

                    num = sci*wht[0].data
                    den = wht[0].data
                    drz_ref = drz
                    drz_ref[0].header['CFILT{0}'.format(i+1)] = utils.parse_filter_from_header(drz[0].header)
                    drz_ref[0].header['NCOMBINE'] = (len(drz_files), 'Number of combined filters')
                else:
                    scl = drz[0].header['PHOTFLAM']/photflam
                    num += sci*scl*(wht[0].data/scl**2)
                    den += wht[0].data/scl**2

                    drz_ref[0].header['CFILT{0}'.format(i+1)] = utils.parse_filter_from_header(drz[0].header)
                    drz_ref[0].header['NDRIZIM'] += drz[0].header['NDRIZIM']

            sci = num/den
            sci[den == 0] = 0
            drz_ref[0].data = sci

            pyfits.writeto('{0}-{1}_drz_sci.fits'.format(label, qfilt),
                           data=sci, header=drz_ref[0].header, overwrite=True,
                           output_verify='fix')

            pyfits.writeto('{0}-{1}_drz_wht.fits'.format(label, qfilt),
                           data=den, header=drz_ref[0].header, overwrite=True,
                           output_verify='fix')

    return grouped_filters


def show_all_thumbnails(label='j022708p4901_00273', filters=['visb', 'visr', 'y', 'j', 'h'], scale_ab=21, close=True, thumb_height=2., rgb_params=RGB_PARAMS, ext='png', xl=0.04, yl=0.98, fs=7):
    """
    Show individual filter and RGB thumbnails
    """
    import glob

    #from PIL import Image

    import numpy as np
    import matplotlib.pyplot as plt

    import astropy.io.fits as pyfits
    from astropy.visualization import make_lupton_rgb
    from grizli.pipeline import auto_script
    from grizli import utils

    all_files = glob.glob('{0}-f*sci.fits'.format(label))
    all_filters = [f.split('_dr')[0].split('-')[-1] for f in all_files]

    ims = {}
    for filter in filters:
        drz_files = glob.glob('{0}-{1}*_dr*sci.fits'.format(label, filter))
        if len(drz_files) > 0:
            im = pyfits.open(drz_files[0])
            ims[filter] = im

    rgb_params['scale_ab'] = scale_ab

    slx, sly, rgb_filts, fig = auto_script.field_rgb(root=label, HOME_PATH=None, **rgb_params)  # xsize=4, output_dpi=None, HOME_PATH=None, show_ir=False, pl=1, pf=1, scl=1, rgb_scl=[1, 1, 1], ds9=None, force_ir=False, filters=all_filters, add_labels=False, output_format='png', rgb_min=-0.01, xyslice=None, pure_sort=False, verbose=True, force_rgb=None, suffix='.rgb', scale_ab=scale_ab)
    if close:
        plt.close()

    #rgb = np.array(Image.open('{0}.rgb.png'.format(label)))
    rgb = plt.imread('{0}.rgb.png'.format(label))

    NX = (len(filters)+1)
    fig = plt.figure(figsize=[thumb_height*NX, thumb_height])
    ax = fig.add_subplot(1, NX, NX)
    ax.imshow(rgb, origin='upper', interpolation='nearest')
    # ax.text(0.05, 0.95, label, ha='left', va='top', transform=ax.transAxes, fontsize=7, color='w', bbox=dict(facecolor='k', edgecolor='None', alpha=0.8))
    # ax.text(0.05, 0.05, ' '.join(rgb_filts), ha='left', va='bottom', transform=ax.transAxes, fontsize=6, color='w', bbox=dict(facecolor='k', edgecolor='None', alpha=0.8))

    for i, filter in enumerate(filters):
        if filter in ims:
            zp_i = utils.calc_header_zeropoint(ims[filter], ext=0)
            scl = 10**(-0.4*(zp_i-5-scale_ab))
            pixscl = utils.get_wcs_pscale(ims[filter][0].header.copy())
            scl *= (0.06/pixscl)**2

            img = ims[filter][0].data*scl

            image = make_lupton_rgb(img, img, img, stretch=0.1, minimum=-0.01)

            ax = fig.add_subplot(1, NX, i+1)
            ax.imshow(255-image, origin='lower', interpolation='nearest')

    for ax in fig.axes:
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])

    fig.tight_layout(pad=0.1)

    # Add labels
    #xl, yl = 0.04, 0.98
    for i, filter in enumerate(filters):
        if filter in ims:
            if filter in ['uv', 'visb', 'visr', 'y', 'j', 'h']:
                grouped_filters = []
                h_i = ims[filter][0].header
                for j in range(h_i['NCOMBINE']):
                    grouped_filters.append(h_i['CFILT{0}'.format(j+1)])

                text_label = '+'.join(grouped_filters)
            else:
                text_label = filter

            fig.text((i+xl)/NX, yl, text_label, fontsize=fs,
                     ha='left', va='top', transform=fig.transFigure,
                     bbox=dict(facecolor='w', edgecolor='None', alpha=0.9))

    fig.text((i+1+xl)/NX, yl, label, ha='left', va='top', transform=fig.transFigure, fontsize=fs, color='w', bbox=dict(facecolor='k', edgecolor='None', alpha=0.8))
    fig.text((i+1+0.04)/NX, 1-yl, ' '.join(rgb_filts), ha='left', va='bottom', transform=fig.transFigure, fontsize=fs, color='w', bbox=dict(facecolor='k', edgecolor='None', alpha=0.8))

    fig.savefig('{0}.thumb.{1}'.format(label, ext))
    if close:
        plt.close()

    return fig


if __name__ == "__main__":
    import sys

    if len(sys.argv) < 5:
        print('Usage: aws_drizzler.py cp561356 150.208875 1.850241667 40 ')
        print(sys.argv)
        exit()

    # print('xxx')
    drizzle_images(label=sys.argv[1], ra=float(sys.argv[2]), dec=float(sys.argv[3]), size=float(sys.argv[4]))
